from __future__ import annotations
from typing import Type, TypeVar
from typing_extensions import Self
import dataclasses
import named_arrays as na

__all__ = [
    'AbstractTemporalSpectralDirectionalVectorArray',
    'TemporalSpectralDirectionalVectorArray',
    'AbstractImplicitTemporalSpectralDirectionalVectorArray',
    'AbstractParameterizedTemporalSpectralDirectionalVectorArray',
    'AbstractTemporalSpectralDirectionalVectorSpace',
    'TemporalSpectralDirectionalVectorLinearSpace',
    'ExplicitTemporalWcsSpectralDirectionalVectorArray',
]

TimeT = TypeVar("TimeT", bound=na.ArrayLike)
DirectionT = TypeVar("DirectionT", bound=na.ArrayLike)
WavelengthT = TypeVar("WavelengthT", bound=na.ScalarLike)


@dataclasses.dataclass(eq=False, repr=False)
class AbstractTemporalSpectralDirectionalVectorArray(
    na.AbstractDirectionalVectorArray,
    na.AbstractSpectralVectorArray,
    na.AbstractTemporalVectorArray,
):

    @property
    def type_abstract(self) -> Type[na.AbstractArray]:
        return AbstractTemporalSpectralDirectionalVectorArray

    @property
    def type_explicit(self) -> Type[na.AbstractExplicitArray]:
        return TemporalSpectralDirectionalVectorArray

    @property
    def type_matrix(self) -> Type[na.TemporalSpectralDirectionalMatrixArray]:
        return na.TemporalSpectralDirectionalMatrixArray


@dataclasses.dataclass(eq=False, repr=False)
class TemporalSpectralDirectionalVectorArray(
    AbstractTemporalSpectralDirectionalVectorArray,
    na.DirectionalVectorArray[DirectionT],
    na.SpectralVectorArray[WavelengthT],
    na.TemporalVectorArray
):

    @classmethod
    def from_scalar(
            cls: Type[Self],
            scalar: na.AbstractScalar,
            like: None | na.AbstractExplicitVectorArray = None,
    ) -> TemporalSpectralDirectionalVectorArray:
        return cls(time=scalar, wavelength=scalar, direction=scalar)


@dataclasses.dataclass(eq=False, repr=False)
class AbstractImplicitTemporalSpectralDirectionalVectorArray(
    AbstractTemporalSpectralDirectionalVectorArray,
    na.AbstractImplicitDirectionalVectorArray,
    na.AbstractImplicitSpectralVectorArray,
    na.AbstractImplicitTemporalVectorArray,
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class AbstractParameterizedTemporalSpectralDirectionalVectorArray(
    AbstractImplicitTemporalSpectralDirectionalVectorArray,
    na.AbstractParameterizedDirectionalVectorArray,
    na.AbstractParameterizedSpectralVectorArray,
    na.AbstractParameterizedTemporalVectorArray,
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class AbstractTemporalSpectralDirectionalVectorSpace(
    AbstractParameterizedTemporalSpectralDirectionalVectorArray,
    na.AbstractDirectionalVectorSpace,
    na.AbstractSpectralVectorSpace,
    na.AbstractTemporalVectorSpace
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class TemporalSpectralDirectionalVectorLinearSpace(
    AbstractTemporalSpectralDirectionalVectorSpace,
    na.AbstractVectorLinearSpace,
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class ExplicitTemporalWcsSpectralDirectionalVectorArray(
    AbstractImplicitTemporalSpectralDirectionalVectorArray,
    na.AbstractWcsVector,
):
    time: na.AbstractScalar = dataclasses.MISSING
    crval: AbstractTemporalSpectralDirectionalVectorArray = dataclasses.MISSING
    crpix: na.AbstractCartesianNdVectorArray = dataclasses.MISSING
    cdelt: AbstractTemporalSpectralDirectionalVectorArray = dataclasses.MISSING
    pc: na.AbstractTemporalSpectralDirectionalMatrixArray = dataclasses.MISSING
    shape_wcs: dict[str, int] = dataclasses.MISSING

    @property
    def _components_explicit(self) -> dict[str, na.ArrayLike]:
        return dict(time=self.time)



from __future__ import annotations
from typing import TypeVar, Type, Generic
from typing_extensions import Self
import abc
import dataclasses
import numpy as np
import astropy.units as u
import named_arrays as na

__all__ = [
    'AbstractCartesian3dVectorArray',
    'Cartesian3dVectorArray',
    'AbstractImplicitCartesian3dVectorArray',
    'AbstractCartesian3dVectorRandomSample',
    'Cartesian3dVectorUniformRandomSample',
    'Cartesian3dVectorNormalRandomSample',
    'AbstractParameterizedCartesian3dVectorArray',
    'Cartesian3dVectorArrayRange',
    'AbstractCartesian3dVectorSpace',
    'Cartesian3dVectorLinearSpace',
    'Cartesian3dVectorStratifiedRandomSpace',
    'Cartesian3dVectorLogarithmicSpace',
    'Cartesian3dVectorGeometricSpace',
]

XT = TypeVar('XT', bound=na.ArrayLike)
YT = TypeVar('YT', bound=na.ArrayLike)
ZT = TypeVar('ZT', bound=na.ArrayLike)


@dataclasses.dataclass(eq=False, repr=False)
class AbstractCartesian3dVectorArray(
    na.AbstractCartesian2dVectorArray,
):

    @property
    @abc.abstractmethod
    def z(self: Self) -> na.ArrayLike:
        """
        The `z` component of the vector.
        """

    @property
    def xy(self) -> na.Cartesian2dVectorArray:
        return na.Cartesian2dVectorArray(
            x=self.x,
            y=self.y,
        )

    @property
    def type_abstract(self: Self) -> Type[AbstractCartesian3dVectorArray]:
        return AbstractCartesian3dVectorArray

    @property
    def type_explicit(self: Self) -> Type[Cartesian3dVectorArray]:
        return Cartesian3dVectorArray

    @property
    def type_matrix(self) -> Type[na.Cartesian3dMatrixArray]:
        return na.Cartesian3dMatrixArray

    @property
    def explicit(self) -> Cartesian3dVectorArray:
        return super().explicit

    def volume_cell(
        self,
        axis: None | tuple[str, str, str],
    ) -> na.AbstractScalar:

        shape = self.shape

        if axis is None:
            axis = tuple(shape)
        else:
            if not set(axis).issubset(shape):
                raise ValueError(
                    f"{axis=} should be a subset of {self.shape=}."
                )

        result = 0

        ax, ay, az = axis

        axes_face = [
            (ax, ay, az),
            (ay, az, ax),
            (az, ax, ay),
        ]

        for axis_face in axes_face:
            a1, a2, a3 = axis_face
            face = [
                self[{a1: slice(None, ~0), a2: slice(None, ~0)}],
                self[{a1: slice(+1, None), a2: slice(None, ~0)}],
                self[{a1: slice(+1, None), a2: slice(+1, None)}],
                self[{a1: slice(None, ~0), a2: slice(+1, None)}],
            ]
            triangles = [
                [face[0], face[1], face[2]],
                [face[2], face[3], face[0]],
            ]
            for triangle in triangles:
                v1, v2, v3 = triangle
                vol = v1 @ v2.cross(v3)
                result = result - vol[{a3: slice(None, ~0)}]
                result = result + vol[{a3: slice(+1, None)}]

        result = result / 6

        return result

    @classmethod
    def _sold_angle(
        cls,
        a: na.AbstractCartesian3dVectorArray,
        b: na.AbstractCartesian3dVectorArray,
        c: na.AbstractCartesian3dVectorArray,
    ) -> na.ScalarLike:

        numerator = a @ b.cross(c)

        a_ = a.length
        b_ = b.length
        c_ = c.length

        d0 = a_ * b_ * c_
        d1 = (a @ b) * c_
        d2 = (a @ c) * b_
        d3 = (b @ c) * a_
        denomerator = d0 + d1 + d2 + d3

        unit = na.unit(numerator)

        if unit is not None:
            numerator = numerator.to(unit).value
            denomerator = denomerator.to(unit).value

        angle = 2 * np.arctan2(numerator, denomerator)

        return angle << u.sr

    def solid_angle_cell(
        self,
        axis: None | tuple[str, str] = None,
    ) -> na.AbstractScalar:
        r"""
        Compute the solid angle of each cell formed by interpreting this
        array as a logically-rectangular 2D grid of vertices.

        Note that this method is usually only used for sorted arrays

        Parameters
        ----------
        axis
            The two axes defining the logically-rectangular 2D grid.
            If :obj:`None` (the default), :attr:`axes` is used and must have
            only two elements.

        Notes
        -----
        The solid angle :math:`\Omega` of a triangle formed by the vertices
        :math:`\vec{a}`, :math:`\vec{b}`, and :math:`\vec{c}` is given by
        :cite:t:`Eriksson1990` as

        .. math::

            \tan \left( \frac{1}{2} \Omega \right)
                = \frac{\vec{a} \cdot (\vec{b} \times \vec{c})}
                    {a b c + (\vec{a} \cdot \vec{b}) c + (\vec{a} \cdot \vec{c}) b + (\vec{b} \cdot \vec{c}) a}.

        Each rectangular cell is decomposed into two triangles and then the
        solid angle of each triangle is computed.
        """

        shape = self.shape

        if axis is None:
            axis = tuple(shape)
        else:
            if not set(axis).issubset(shape):   # pragma: nocover
                raise ValueError(
                    f"{axis=} should be a subset of {self.shape=}."
                )

        ax, ay = axis

        s0 = slice(None, ~0)
        s1 = slice(+1, None)

        a = self[{ax: s0, ay: s0}]
        b = self[{ax: s1, ay: s0}]
        c = self[{ax: s1, ay: s1}]
        d = self[{ax: s0, ay: s1}]

        angle_1 = self._sold_angle(a, b, c)
        angle_2 = self._sold_angle(c, d, a)

        return angle_1 + angle_2

    def cross(
            self,
            other: AbstractCartesian3dVectorArray,
    ) -> na.Cartesian3dVectorArray:
        r"""
        Compute the vector product :math:`\mathbf{a} \times \mathbf{b}`

        Parameters
        ----------
        other
            the right-hand operand of the cross product operation
        """
        a = self
        b = other
        if isinstance(other, na.AbstractCartesian3dVectorArray):
            if len(self.cartesian_nd.components) != 3:
                raise ValueError("all components of `self` must be scalars")
            if len(other.cartesian_nd.components) != 3:
                raise ValueError("all components of `other` must be scalars")
            return self.type_explicit(
                x=+(a.y * b.z - a.z * b.y),
                y=-(a.x * b.z - a.z * b.x),
                z=+(a.x * b.y - a.y * b.x),
            )
        else:
            raise TypeError(
                f"`other` must be an instance of `{na.AbstractCartesian3dVectorArray.__name__}`, "
                f"got `{type(other).__name__}`"
            )


@dataclasses.dataclass(eq=False, repr=False)
class Cartesian3dVectorArray(
    AbstractCartesian3dVectorArray,
    na.Cartesian2dVectorArray,
    Generic[XT, YT, ZT],
):
    x: XT = 0
    y: YT = 0
    z: ZT = 0

    @classmethod
    def from_scalar(
            cls: Type[Self],
            scalar: na.AbstractScalar,
            like: None | na.AbstractExplicitVectorArray = None,
    ) -> Cartesian3dVectorArray:
        result = super().from_scalar(scalar, like=like)
        result.z = scalar
        return result


@dataclasses.dataclass(eq=False, repr=False)
class AbstractImplicitCartesian3dVectorArray(
    AbstractCartesian3dVectorArray,
    na.AbstractImplicitCartesian2dVectorArray,
):

    @property
    def x(self) -> na.ArrayLike:
        return self.explicit.x

    @property
    def y(self) -> na.ArrayLike:
        return self.explicit.y

    @property
    def z(self) -> na.ArrayLike:
        return self.explicit.z


@dataclasses.dataclass(eq=False, repr=False)
class AbstractCartesian3dVectorRandomSample(
    AbstractImplicitCartesian3dVectorArray,
    na.AbstractCartesian2dVectorRandomSample,
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class Cartesian3dVectorUniformRandomSample(
    AbstractCartesian3dVectorRandomSample,
    na.Cartesian2dVectorUniformRandomSample,
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class Cartesian3dVectorNormalRandomSample(
    AbstractCartesian3dVectorRandomSample,
    na.Cartesian2dVectorNormalRandomSample,
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class AbstractParameterizedCartesian3dVectorArray(
    AbstractImplicitCartesian3dVectorArray,
    na.AbstractParameterizedCartesian2dVectorArray,
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class Cartesian3dVectorArrayRange(
    AbstractParameterizedCartesian3dVectorArray,
    na.Cartesian2dVectorArrayRange,
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class AbstractCartesian3dVectorSpace(
    AbstractParameterizedCartesian3dVectorArray,
    na.AbstractCartesian2dVectorSpace,
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class Cartesian3dVectorLinearSpace(
    AbstractCartesian3dVectorSpace,
    na.Cartesian2dVectorLinearSpace,
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class Cartesian3dVectorStratifiedRandomSpace(
    Cartesian3dVectorLinearSpace,
    na.Cartesian2dVectorStratifiedRandomSpace,
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class Cartesian3dVectorLogarithmicSpace(
    AbstractCartesian3dVectorSpace,
    na.Cartesian2dVectorLogarithmicSpace,
):
    pass


@dataclasses.dataclass(eq=False, repr=False)
class Cartesian3dVectorGeometricSpace(
    AbstractCartesian3dVectorSpace,
    na.Cartesian2dVectorGeometricSpace,
):
    pass



from typing import Type, Sequence, Callable
import numpy as np
import pytest
import astropy.units as u
import named_arrays as na
import named_arrays.tests.test_core
import named_arrays._scalars.tests.test_scalars

__all__ = [
    'AbstractTestAbstractUncertainScalarArray',
    'TestUncertainScalarArray',
    'TestUncertainScalarArrayCreation',
    'AbstractTestAbstractImplicitUncertainScalarArray',
    'TestUniformUncertainScalarArray',
    'TestNormalUncertainScalarArray',
]

_num_x = named_arrays.tests.test_core.num_x
_num_y = named_arrays.tests.test_core.num_y
_num_distribution = named_arrays.tests.test_core.num_distribution


def _uncertain_scalar_arrays():
    arrays_nominal = [
        4.,
        na.ScalarArray(4.),
        na.ScalarUniformRandomSample(-4, 4, shape_random=dict(x=_num_x, y=_num_y))
    ]
    arrays_distribution = [
        4 + na.ScalarUniformRandomSample(-0.1, 0.1, shape_random=dict(_distribution=_num_distribution)),
        4 + na.ScalarUniformRandomSample(
            start=-0.1,
            stop=0.1,
            shape_random=dict(x=_num_x, y=_num_y, _distribution=_num_distribution)
        ),
    ]
    units = [1, u.mm]
    arrays = [
        na.UncertainScalarArray(nominal * unit, distribution * unit)
        for nominal in arrays_nominal
        for distribution in arrays_distribution
        for unit in units
    ]
    return arrays


def _uncertain_scalar_arrays_2():
    arrays_exact = [
        5,
        na.ScalarUniformRandomSample(-5, 5, shape_random=dict(y=_num_y)),
    ]
    arrays_nominal = arrays_exact
    arrays_distribution = [
        na.ScalarArray(5.1).add_axes(na.UncertainScalarArray.axis_distribution),
        5 + na.ScalarUniformRandomSample(
            start=-5.1,
            stop=5.1,
            shape_random=dict(x=_num_x, y=_num_y, _distribution=_num_distribution),
        ),
    ]
    units = [1, u.mm]
    arrays_uncertain = [
        na.UncertainScalarArray(nominal * unit, distribution * unit)
        for nominal in arrays_nominal
        for distribution in arrays_distribution
        for unit in units
    ]
    arrays = arrays_exact + arrays_uncertain
    return arrays


class AbstractTestAbstractUncertainScalarArray(
    named_arrays._scalars.tests.test_scalars.AbstractTestAbstractScalar,
):

    def test_nominal(self, array: na.AbstractUncertainScalarArray):
        assert np.sum(array.nominal) != 0
        assert array.axis_distribution not in na.shape(array.nominal)

    def test_distribution(self, array: na.AbstractUncertainScalarArray):
        assert np.sum(array.distribution) != 0

    def test_num_distribution(self, array: na.AbstractUncertainScalarArray):
        assert isinstance(array.num_distribution, int)
        assert array.num_distribution > 0

    def test_shape_distribution(self, array: na.AbstractUncertainScalarArray):
        assert isinstance(array.shape_distribution, dict)
        for ax in array.shape_distribution:
            assert isinstance(ax, str)
            assert isinstance(array.shape_distribution[ax], int)

        assert array.axis_distribution in array.shape_distribution

    @pytest.mark.parametrize(
        argnames='item',
        argvalues=[
            dict(y=0),
            dict(y=slice(0, 1)),
            dict(y=na.ScalarArray(np.array([0, 1]), axes=('y', ))),
            dict(
                y=na.UncertainScalarArray(
                    nominal=na.ScalarArray(np.array([0, 1]), axes=('y', )),
                    distribution=na.ScalarArray(
                        ndarray=np.array([[0,], [1,]]),
                        axes=('y', na.UncertainScalarArray.axis_distribution),
                    )
                ),
                _distribution=na.UncertainScalarArray(
                    nominal=None,
                    distribution=na.ScalarArray(
                        ndarray=np.array([[0], [0]]),
                        axes=('y', na.UncertainScalarArray.axis_distribution),
                    )
                )
            ),
            na.ScalarLinearSpace(0, 1, axis='y', num=_num_y) > 0.5,
            na.UncertainScalarArray(
                nominal=na.ScalarLinearSpace(0, 1, axis='y', num=_num_y),
                distribution=na.ScalarNormalRandomSample(
                    center=na.ScalarLinearSpace(0, 1, axis='y', num=_num_y),
                    width=0.1,
                    shape_random={na.UncertainScalarArray.axis_distribution: _num_distribution},
                )
            ) > 0.5,
        ]
    )
    def test__getitem__(
            self,
            array: na.AbstractUncertainScalarArray,
            item: dict[str, int | slice | na.AbstractArray] | na.AbstractArray
    ):
        super().test__getitem__(array=array, item=item)

        if isinstance(item, na.AbstractArray):

            if not set(item.shape).issubset(array.shape_distribution):
                with pytest.raises(ValueError):
                    array[item]
                return

            if isinstance(item, na.AbstractUncertainScalarArray):
                item_nominal = item.nominal & np.all(item.distribution, na.UncertainScalarArray.axis_distribution)
                item_distribution = item_nominal
            else:
                item_nominal = item_distribution = item

        elif isinstance(item, dict):

            if not set(item).issubset(array.shape_distribution):
                with pytest.raises(ValueError):
                    array[item]
                return

            item_nominal = dict()
            item_distribution = dict()

            for ax in item:
                if isinstance(item[ax], na.AbstractArray):
                    if isinstance(item[ax], na.AbstractUncertainScalarArray):
                        item_nominal[ax] = item[ax].nominal
                        item_distribution[ax] = item[ax].distribution
                    else:
                        item_nominal[ax] = item_distribution[ax] = item[ax]
                else:
                    item_nominal[ax] = item_distribution[ax] = item[ax]

        result = array[item]
        result_expected = na.UncertainScalarArray(
            array.broadcasted.nominal[item_nominal],
            array.broadcasted.distribution[item_distribution],
        )

        assert np.all(result == result_expected)

    def test__mul__(self, array: na.AbstractUncertainScalarArray):
        unit = u.mm
        result = array * unit
        result_nominal = array.nominal * unit
        result_distribution = array.distribution * unit
        assert np.all(result.nominal == result_nominal)
        assert np.all(result.distribution == result_distribution)

    def test__lshift__(self, array: na.AbstractUncertainScalarArray):
        unit = u.mm
        result = array << unit
        result_nominal = array.nominal << unit
        result_distribution = array.distribution << unit
        assert np.all(result.nominal == result_nominal)
        assert np.all(result.distribution == result_distribution)

    def test__truediv__(self, array: na.AbstractUncertainScalarArray):
        unit = u.mm
        result = array / unit
        result_nominal = array.nominal / unit
        result_distribution = array.distribution / unit
        assert np.all(result.nominal == result_nominal)
        assert np.all(result.distribution == result_distribution)

    class TestUfuncUnary(
        named_arrays._scalars.tests.test_scalars.AbstractTestAbstractScalar.TestUfuncUnary
    ):

        def test_ufunc_unary(
                self,
                ufunc: np.ufunc,
                array: na.AbstractUncertainScalarArray,
        ):
            super().test_ufunc_unary(ufunc, array)

            kwargs = dict()
            kwargs_nominal = dict()
            kwargs_distribution = dict()

            unit = array.unit_normalized
            if ufunc in [np.log, np.log2, np.log10, np.sqrt]:
                kwargs["where"] = array > 0
            elif ufunc in [np.log1p]:
                kwargs["where"] = array >= (-1 * unit)
            elif ufunc in [np.arcsin, np.arccos, np.arctanh]:
                kwargs["where"] = ((-1 * unit) < array) & (array < (1 * unit))
            elif ufunc in [np.arccosh]:
                kwargs["where"] = array >= (1 * unit)
            elif ufunc in [np.reciprocal]:
                kwargs["where"] = array != 0

            if "where" in kwargs:
                kwargs_nominal["where"] = kwargs["where"].nominal
                kwargs_distribution["where"] = kwargs["where"].distribution

            try:
                ufunc(array.nominal, **kwargs_nominal)
                ufunc(array.distribution, **kwargs_distribution)
            except (ValueError, TypeError) as e:
                with pytest.raises(type(e)):
                    ufunc(array, **kwargs)
                return

            result = ufunc(array, **kwargs)
            result_nominal = ufunc(array.nominal, **kwargs_nominal)
            result_distribution = ufunc(array.distribution, **kwargs_distribution)

            if ufunc.nout == 1:
                out = 0 * result
            else:
                out = tuple(0 * r for r in result)

            result_out = ufunc(array, out=out, **kwargs)

            if ufunc.nout == 1:
                out = (out, )
                result = (result, )
                result_nominal = (result_nominal, )
                result_distribution = (result_distribution, )
                result_out = (result_out, )

            for i in range(ufunc.nout):
                assert np.all(result[i].nominal == result_nominal[i], **kwargs_nominal)
                assert np.all(result[i].distribution == result_distribution[i], **kwargs_distribution)
                assert np.all(result[i] == result_out[i], **kwargs)
                assert result_out[i] is out[i]

    @pytest.mark.parametrize('array_2', _uncertain_scalar_arrays_2())
    class TestUfuncBinary(
        named_arrays._scalars.tests.test_scalars.AbstractTestAbstractScalar.TestUfuncBinary
    ):

        def test_ufunc_binary(
                self,
                ufunc: np.ufunc,
                array: None | bool | int | float | complex | na.AbstractUncertainScalarArray,
                array_2: None | bool | int | float | complex | na.AbstractUncertainScalarArray,
        ):
            super().test_ufunc_binary(ufunc=ufunc, array=array, array_2=array_2)

            if not isinstance(array, na.AbstractUncertainScalarArray):
                array_normalized = na.UncertainScalarArray(
                    nominal=array,
                    distribution=na.add_axes(array, na.UncertainScalarArray.axis_distribution),
                )
            else:
                array_normalized = array

            if not isinstance(array_2, na.AbstractUncertainScalarArray):
                array_2_normalized = na.UncertainScalarArray(
                    nominal=array_2,
                    distribution=na.add_axes(array_2, na.UncertainScalarArray.axis_distribution),
                )
            else:
                array_2_normalized = array_2

            kwargs = dict()
            kwargs_nominal = dict()
            kwargs_distribution = dict()

            unit_2 = na.unit_normalized(array_2)
            if ufunc in [np.power, np.float_power]:
                kwargs["where"] = (array_2_normalized >= (1 * unit_2)) & (array_normalized >= 0)
            elif ufunc in [np.divide, np.floor_divide, np.remainder, np.fmod, np.divmod]:
                kwargs["where"] = array_2_normalized != 0

            if "where" in kwargs:
                kwargs_nominal["where"] = kwargs["where"].nominal
                kwargs_distribution["where"] = kwargs["where"].distribution

            try:
                ufunc(array_normalized.nominal, array_2_normalized.nominal, **kwargs_nominal)
                ufunc(array_normalized.distribution, array_2_normalized.distribution, **kwargs_distribution)
            except (ValueError, TypeError) as e:
                with pytest.raises(type(e)):
                    ufunc(array, array_2, **kwargs)
                return

            result = ufunc(array, array_2, **kwargs)
            result_nominal = ufunc(array_normalized.nominal, array_2_normalized.nominal, **kwargs_nominal)
            result_distribution = ufunc(
                array_normalized.distribution, array_2_normalized.distribution, **kwargs_distribution)

            if ufunc.nout == 1:
                out = 0 * np.nan_to_num(result)
            else:
                out = tuple(0 * np.nan_to_num(r) for r in result)

            result_out = ufunc(array, array_2, out=out, **kwargs)

            if ufunc.nout == 1:
                out = (out, )
                result = (result, )
                result_nominal = (result_nominal, )
                result_distribution = (result_distribution, )
                result_out = (result_out, )

            for i in range(ufunc.nout):
                assert np.all(result[i].nominal == result_nominal[i], **kwargs_nominal)
                assert np.all(result[i].distribution == result_distribution[i], **kwargs_distribution)
                assert np.all(result[i] == result_out[i], **kwargs)
                assert result_out[i] is out[i]

    @pytest.mark.parametrize('array_2', _uncertain_scalar_arrays_2())
    class TestMatmul(
        named_arrays._scalars.tests.test_scalars.AbstractTestAbstractScalar.TestMatmul
    ):
        pass

    class TestArrayFunctions(
        named_arrays._scalars.tests.test_scalars.AbstractTestAbstractScalar.TestArrayFunctions,
    ):

        @pytest.mark.parametrize("array_2", _uncertain_scalar_arrays_2())
        class TestAsArrayLikeFunctions(
            named_arrays._scalars.tests.test_scalars.AbstractTestAbstractScalar.TestArrayFunctions
            .TestAsArrayLikeFunctions
        ):

            def test_asarray_like_functions(
                    self,
                    func: Callable,
                    array: None | float | u.Quantity | na.AbstractArray,
                    array_2: None | float | u.Quantity | na.AbstractArray,
            ):
                a = array
                like = array_2

                if a is None:
                    assert func(a, like=like) is None
                    return

                result = func(a, like=like)

                assert isinstance(result, na.UncertainScalarArray)
                assert isinstance(result.nominal, na.ScalarArray)
                assert isinstance(result.distribution, na.ScalarArray)
                assert isinstance(result.nominal.ndarray, np.ndarray)
                assert isinstance(result.distribution.ndarray, np.ndarray)

                assert np.all(result.value == na.value(a))

                super().test_asarray_like_functions(
                    func=func,
                    array=array,
                    array_2=array_2,
                )

        class TestSingleArgumentFunctions(
            named_arrays._scalars.tests.test_scalars.AbstractTestAbstractScalar.TestArrayFunctions.TestSingleArgumentFunctions,
        ):
            def test_single_argument_functions(
                self,
                func: Callable,
                array: na.AbstractUncertainScalarArray,
            ):
                result = func(array)
                assert np.all(result.nominal == func(array.nominal))
                assert np.all(result.distribution == func(array.distribution))

        @pytest.mark.parametrize(
            argnames='where',
            argvalues=[
                np._NoValue,
                True,
                na.ScalarArray(True),
                (na.ScalarLinearSpace(-1, 1, 'x', _num_x) >= 0) | (na.ScalarLinearSpace(-1, 1, 'y', _num_y) >= 0),
                na.UncertainScalarArray(
                    nominal=(na.ScalarLinearSpace(-1, 1, 'x', _num_x) >= 0)
                            | (na.ScalarLinearSpace(-1, 1, 'y', _num_y) >= 0),
                    distribution=(na.ScalarLinearSpace(-1, 1, 'x', _num_x) >= 0)
                                 | (na.ScalarLinearSpace(-1, 1, 'y', _num_y) >= 0)
                                 | (na.ScalarLinearSpace(-1, 1, '_distribution', _num_distribution) >= 0)
                ),
            ]
        )
        class TestReductionFunctions(
            named_arrays._scalars.tests.test_scalars.AbstractTestAbstractScalar.TestArrayFunctions.
            TestReductionFunctions,
        ):

            def test_reduction_functions(
                    self,
                    func: Callable,
                    array: na.AbstractUncertainScalarArray,
                    axis: None | str | Sequence[str],
                    dtype: None | type | np.dtype,
                    keepdims: bool,
                    where: bool | na.AbstractArray,
            ):
                super().test_reduction_functions(
                    func=func,
                    array=array,
                    axis=axis,
                    dtype=dtype,
                    keepdims=keepdims,
                    where=where,
                )

                kwargs = dict(
                    axis=axis,
                    dtype=dtype,
                    keepdims=keepdims,
                    where=where,
                )

                kwargs_nominal = kwargs.copy()
                kwargs_distribution = kwargs.copy()

                if axis is None:
                    axis_normalized = na.axis_normalized(array, axis)
                    kwargs_nominal["axis"] = axis_normalized
                    kwargs_distribution["axis"] = axis_normalized

                if isinstance(where, na.AbstractUncertainScalarArray):
                    kwargs_nominal["where"] = where.nominal
                    kwargs_distribution["where"] = where.distribution
                else:
                    kwargs_nominal["where"] = kwargs_distribution["where"] = where

                try:
                    result_nominal = func(array.broadcasted.nominal, **kwargs_nominal)
                    result_distribution = func(array.broadcasted.distribution, **kwargs_distribution)
                except (ValueError, TypeError, u.UnitsError) as e:
                    with pytest.raises(type(e)):
                        func(array, **kwargs)
                    return

                result = func(array, **kwargs)

                out = 0 * result
                result_out = func(array, out=out, **kwargs)

                assert np.all(result.nominal == result_nominal)
                assert np.all(result.distribution == result_distribution)
                assert np.allclose(result, result_out)
                assert result_out is out

        @pytest.mark.parametrize(
            argnames='q',
            argvalues=[
                .25,
                25 * u.percent,
                na.ScalarLinearSpace(.25, .75, axis='q', num=3, endpoint=True),
                na.UncertainScalarArray(
                    nominal=25 * u.percent,
                    distribution=na.ScalarNormalRandomSample(
                        center=25 * u.percent,
                        width=1 * u.percent,
                        shape_random=dict(_distribution=_num_distribution)
                    )
                )
            ]
        )
        class TestPercentileLikeFunctions(
            named_arrays._scalars.tests.test_scalars.AbstractTestAbstractScalar.TestArrayFunctions.
            TestPercentileLikeFunctions
        ):

            def test_percentile_like_functions(
                    self,
                    func: Callable,
                    array: na.AbstractUncertainScalarArray,
                    q: float | u.Quantity | na.AbstractArray,
                    axis: None | str | Sequence[str],
                    keepdims: bool,
            ):
                super().test_percentile_like_functions(
                    func=func,
                    array=array,
                    q=q,
                    axis=axis,
                    keepdims=keepdims,
                )

                array_normalized = array
                if isinstance(array, na.AbstractArray):
                    if isinstance(array, na.AbstractScalar):
                        if isinstance(array, na.AbstractScalarArray):
                            array_normalized = na.UncertainScalarArray(array, array)

                kwargs = dict(
                    q=q,
                    axis=axis,
                    keepdims=keepdims,
                )

                kwargs_nominal = kwargs.copy()
                kwargs_distribution = kwargs.copy()

                if isinstance(q, na.AbstractUncertainScalarArray):
                    kwargs_nominal["q"] = q.nominal
                    kwargs_distribution["q"] = q.distribution
                else:
                    kwargs_nominal["q"] = kwargs_distribution["q"] = q

                if axis is None:
                    axis_normalized = na.axis_normalized(array, axis)
                    kwargs_nominal["axis"] = axis_normalized
                    kwargs_distribution["axis"] = axis_normalized

                try:
                    result_nominal = func(array_normalized.broadcasted.nominal, **kwargs_nominal)
                    result_distribution = func(array_normalized.broadcasted.distribution, **kwargs_distribution)
                except (ValueError, TypeError) as e:
                    with pytest.raises(type(e)):
                        func(array, **kwargs)
                    return

                result = func(array, **kwargs)

                out = 0 * result
                result_out = func(array, out=out, **kwargs)

                assert np.all(result.nominal == result_nominal)
                assert np.all(result.distribution == result_distribution)
                assert np.all(result == result_out)
                assert result_out is out

        class TestFFTLikeFunctions(
            named_arrays._scalars.tests.test_scalars.AbstractTestAbstractScalar.TestArrayFunctions.TestFFTLikeFunctions,
        ):

            def test_fft_like_functions(
                    self,
                    func: Callable,
                    array: na.AbstractUncertainScalarArray,
                    axis: tuple[str, str],
            ):
                if axis[0] not in array.shape:
                    with pytest.raises(ValueError, match="`axis` .* not in array with shape .*"):
                        func(array, axis=axis)
                    return

                result = func(array, axis=axis)
                result_nominal = func(array.broadcasted.nominal, axis=axis)
                result_distribution = func(array.broadcasted.distribution, axis=axis)

                assert np.all(result.nominal == result_nominal)
                assert np.all(result.distribution == result_distribution)

        class TestFFTNLikeFunctions(
            named_arrays._scalars.tests.test_scalars.AbstractTestAbstractScalar.TestArrayFunctions.TestFFTNLikeFunctions
        ):
            def test_fftn_like_functions(
                    self,
                    func: Callable,
                    array: na.AbstractUncertainScalarArray,
                    axes: dict[str, str],
                    s: None | dict[str, int],
            ):
                if not set(axes).issubset(array.shape):
                    with pytest.raises(ValueError, match="`axes`, .*, not a subset of array axes, .*"):
                        func(array, axes=axes, s=s)
                    return

                if s is not None and axes.keys() != s.keys():
                    with pytest.raises(ValueError):
                        func(a=array, axes=axes, s=s)
                    return

                result = func(array, axes=axes, s=s)
                result_nominal = func(array.broadcasted.nominal, axes=axes, s=s)
                result_distribution = func(array.broadcasted.distribution, axes=axes, s=s)

                assert np.all(result.nominal == result_nominal)
                assert np.all(result.distribution == result_distribution)

        class TestEmathFunctions(
            named_arrays._scalars.tests.test_scalars.AbstractTestAbstractScalar.TestArrayFunctions.TestEmathFunctions,
        ):
            def test_emath_functions(
                self,
                func: Callable,
                array: na.AbstractUncertainScalarArray,
            ):
                result = func(array)
                assert np.all(result.nominal == func(array.nominal))
                assert np.all(result.distribution == func(array.distribution))

        @pytest.mark.parametrize('axis', [None, 'x', 'y', ('x', 'y'), ()])
        def test_sort(self, array: na.AbstractUncertainScalarArray, axis: None | str | Sequence[str]):

            axis_normalized = na.axis_normalized(array, axis)

            if axis is not None:
                if not axis:
                    with pytest.raises(ValueError, match=f"if `axis` is a sequence, it must not be empty, got .*"):
                        np.sort(array, axis=axis)
                    return

                if not set(axis_normalized).issubset(array.shape):
                    with pytest.raises(ValueError, match="`axis`, .* is not a subset of `a.axes`, .*"):
                        np.sort(array, axis=axis)
                    return

            result = np.sort(array, axis=axis)

            array_broadcasted = na.broadcast_to(array, array.shape)
            if axis_normalized:
                result_distribution = np.sort(
                    a=array_broadcasted.distribution.mean(array.axis_distribution),
                    axis=axis_normalized,
                )
            else:
                result_distribution = array.distribution.mean(array.axis_distribution)

            assert np.all(result.distribution.mean(array.axis_distribution) == result_distribution)

        @pytest.mark.parametrize('copy', [False, True])
        def test_nan_to_num(
                self,
                array: na.AbstractUncertainScalarArray,
                copy: bool,
        ):

            super().test_nan_to_num(array=array, copy=copy)

            if not copy and not isinstance(array, na.AbstractExplicitArray):
                with pytest.raises(TypeError, match="can't write to an array .*"):
                    np.nan_to_num(array, copy=copy)
                return

            result = np.nan_to_num(array, copy=copy)
            result_nominal = np.nan_to_num(array.nominal, copy=copy)
            result_distribution = np.nan_to_num(array.distribution, copy=copy)

            assert np.all(result.nominal == result_nominal)
            assert np.all(result.distribution == result_distribution)

        @pytest.mark.parametrize('v', _uncertain_scalar_arrays_2())
        @pytest.mark.parametrize('mode', ['full', 'same', 'valid'])
        def test_convolve(self, array: na.AbstractArray, v: na.AbstractArray, mode: str):
            super().test_convolve(array=array, v=v, mode=mode)
            with pytest.raises(ValueError, match="`numpy.convolve` is not supported .*"):
                np.convolve(array, v=v, mode=mode)

    class TestNamedArrayFunctions(
        named_arrays._scalars.tests.test_scalars.AbstractTestAbstractScalar.TestNamedArrayFunctions
    ):
        @pytest.mark.parametrize(
            argnames="array_2",
            argvalues=_uncertain_scalar_arrays_2(),
        )
        @pytest.mark.parametrize(
            argnames="where",
            argvalues=[
                np._NoValue,
                True,
                na.linspace(0, 1, axis="x", num=_num_x) > 0.5,
            ]
        )
        @pytest.mark.parametrize(
            argnames="alpha",
            argvalues=[
                np._NoValue,
                na.linspace(0, 1, axis="x", num=_num_x),
            ]
        )
        class TestPltPlotLikeFunctions(
            named_arrays._scalars.tests.test_scalars.AbstractTestAbstractScalar.TestNamedArrayFunctions.TestPltPlotLikeFunctions,
        ):
            pass

        @pytest.mark.parametrize(
            argnames="array_2",
            argvalues=_uncertain_scalar_arrays_2(),
        )
        @pytest.mark.parametrize(
            argnames="s",
            argvalues=[
                None,
            ]
        )
        @pytest.mark.parametrize(
            argnames="c",
            argvalues=[
                None,
            ]
        )
        @pytest.mark.parametrize(
            argnames="where",
            argvalues=[
                True,
                na.linspace(0, 1, axis="x", num=_num_x) > 0.5,
            ]
        )
        class TestPltScatter(
            named_arrays._scalars.tests.test_scalars.AbstractTestAbstractScalar.TestNamedArrayFunctions.TestPltScatter,
        ):
            pass

        @pytest.mark.skip
        class TestPltPcolormesh(
            named_arrays._scalars.tests.test_scalars.AbstractTestAbstractScalar.TestNamedArrayFunctions.TestPltPcolormesh,
        ):
            pass

        @pytest.mark.parametrize(
            argnames="function",
            argvalues=[
                lambda x: a * x ** 3
                for a in [2, na.UniformUncertainScalarArray(2, width=0.5, num_distribution=_num_distribution)]
            ]
        )
        class TestJacobian(
            named_arrays._scalars.tests.test_scalars.AbstractTestAbstractScalar.TestNamedArrayFunctions.TestJacobian,
        ):
            pass

        @pytest.mark.parametrize(
            argnames="func",
            argvalues=[
                na.optimize.root_secant,
                na.optimize.root_newton,
            ],
        )
        @pytest.mark.parametrize(
            argnames="function",
            argvalues=[
                lambda x: np.square(na.value(x) - shift_horizontal) + shift_vertical
                for shift_horizontal in [
                    20,
                    na.UniformUncertainScalarArray(20, width=1, num_distribution=_num_distribution),
                ]
                for shift_vertical in [-1]
            ]
        )
        class TestOptimizeRoot(
            named_arrays._scalars.tests.test_scalars.AbstractTestAbstractScalar.TestNamedArrayFunctions.TestOptimizeRoot,
        ):
            pass

        @pytest.mark.skip
        class TestOptimizeMinimum(
            named_arrays._scalars.tests.test_scalars.AbstractTestAbstractScalar.TestNamedArrayFunctions.TestOptimizeMinimum,
        ):
            pass


@pytest.mark.parametrize('array', _uncertain_scalar_arrays())
class TestUncertainScalarArray(
    AbstractTestAbstractUncertainScalarArray,
    named_arrays.tests.test_core.AbstractTestAbstractExplicitArray
):

    @pytest.mark.parametrize(
        argnames="item",
        argvalues=[
            dict(y=0),
            dict(x=0, y=0),
            dict(y=slice(None)),
            dict(y=na.ScalarArrayRange(0, _num_y, axis='y')),
            dict(x=na.ScalarArrayRange(0, _num_x, axis='x'), y=na.ScalarArrayRange(0, _num_y, axis='y')),
            na.ScalarArray.ones(shape=dict(y=_num_y), dtype=bool),
        ],
    )
    @pytest.mark.parametrize(
        argnames="value",
        argvalues=[
            0,
            na.ScalarUniformRandomSample(-5, 5, dict(y=_num_y)),
            na.UncertainScalarUniformRandomSample(
                start=na.UniformUncertainScalarArray(-5, 0, num_distribution=_num_distribution),
                stop=5,
                shape_random=dict(y=_num_y),
            )
        ]
    )
    def test__setitem__(
            self,
            array: na.ScalarArray,
            item: dict[str, int | slice | na.ScalarArray] | na.ScalarArray,
            value: float | na.ScalarArray
    ):
        super().test__setitem__(array=array, item=item, value=value)


@pytest.mark.parametrize("type_array", [na.UncertainScalarArray])
class TestUncertainScalarArrayCreation(
    named_arrays.tests.test_core.AbstractTestAbstractExplicitArrayCreation,
):

    @pytest.mark.parametrize("like", [None] + _uncertain_scalar_arrays())
    class TestFromScalarArray(
        named_arrays.tests.test_core.AbstractTestAbstractExplicitArrayCreation.TestFromScalarArray,
    ):
        pass


class AbstractTestAbstractImplicitUncertainScalarArray(
    AbstractTestAbstractUncertainScalarArray,
    named_arrays.tests.test_core.AbstractTestAbstractImplicitArray,
):
    pass


def _uniform_uncertain_scalar_arrays():
    arrays_exact = [
        4,
        na.ScalarUniformRandomSample(-4, 4, shape_random=dict(x=_num_x, y=_num_y)),
    ]
    widths = [
        1,
        na.ScalarLinearSpace(1, 2, axis='y', num=_num_y)
    ]
    units = [
        1,
    ]
    arrays = [
        na.UniformUncertainScalarArray(
            nominal=array_exact * unit,
            width=width * unit,
            num_distribution=_num_distribution
        )
        for array_exact in arrays_exact
        for width in widths
        for unit in units
    ]
    return arrays


@pytest.mark.parametrize('array', _uniform_uncertain_scalar_arrays())
class TestUniformUncertainScalarArray(
    AbstractTestAbstractImplicitUncertainScalarArray,
):
    pass


def _normal_uncertain_scalar_arrays():
    arrays_exact = [
        na.ScalarArray(4),
        na.ScalarUniformRandomSample(-4, 4, shape_random=dict(x=_num_x, y=_num_y)),
    ]
    widths = [
        1,
        na.ScalarLinearSpace(1, 2, axis='y', num=_num_y)
    ]
    units = [
        u.mm,
    ]
    arrays = [
        na.NormalUncertainScalarArray(
            nominal=array_exact * unit,
            width=width * unit,
            num_distribution=_num_distribution
        )
        for array_exact in arrays_exact
        for width in widths
        for unit in units
    ]
    return arrays


@pytest.mark.parametrize('array', _normal_uncertain_scalar_arrays())
class TestNormalUncertainScalarArray(
    AbstractTestAbstractImplicitUncertainScalarArray,
):
    pass


class AbstractTestAbstractUncertainScalarRandomSample(
    AbstractTestAbstractImplicitUncertainScalarArray,
    named_arrays.tests.test_core.AbstractTestAbstractRandomSample,
):
    pass


def _uncertain_scalar_uniform_random_samples() -> tuple[na.UncertainScalarUniformRandomSample, ...]:

    starts = (
        na.UniformUncertainScalarArray(
            na.ScalarLinearSpace(0, 1, axis='x', num=_num_x),
            width=0.2,
            num_distribution=_num_distribution,
        ),
    )
    stops = (
        2,
        na.UniformUncertainScalarArray(
            na.ScalarLinearSpace(2, 3, axis='x', num=_num_x),
            width=0.1,
            num_distribution=_num_distribution,
        ),
    )
    units = [
        u.mm,
    ]
    arrays = tuple(
        na.UncertainScalarUniformRandomSample(
            start=start * unit,
            stop=stop * unit,
            shape_random=dict(y=_num_y),
        )
        for start in starts
        for stop in stops
        for unit in units
    )
    return arrays


@pytest.mark.parametrize("array", _uncertain_scalar_uniform_random_samples())
class TestUncertainScalarUniformRandomSample(
    AbstractTestAbstractUncertainScalarRandomSample,
    named_arrays.tests.test_core.AbstractTestAbstractUniformRandomSample,
):
    pass


def _uncertain_scalar_normal_random_samples() -> tuple[na.UncertainScalarNormalRandomSample, ...]:

    centers = (
        na.UniformUncertainScalarArray(
            na.ScalarLinearSpace(0, 1, axis='x', num=_num_x),
            width=0.2,
            num_distribution=_num_distribution,
        ),
    )
    widths = (
        na.ScalarLinearSpace(2, 3, axis='x', num=_num_x),
        na.UniformUncertainScalarArray(
            na.ScalarLinearSpace(2, 3, axis='x', num=_num_x),
            width=0.1,
            num_distribution=_num_distribution,
        ),
    )
    units = [
        1,
    ]
    arrays = tuple(
        na.UncertainScalarNormalRandomSample(
            center=center * unit,
            width=width * unit,
            shape_random=dict(y=_num_y),
        )
        for center in centers
        for width in widths
        for unit in units
    )
    return arrays


@pytest.mark.parametrize("array", _uncertain_scalar_normal_random_samples())
class TestUncertainScalarNormalRandomSample(
    AbstractTestAbstractUncertainScalarRandomSample,
    named_arrays.tests.test_core.AbstractTestAbstractNormalRandomSample,
):
    pass


def _uncertain_scalar_poisson_random_samples() -> tuple[na.UncertainScalarPoissionRandomSample, ...]:

    centers = (
        na.UniformUncertainScalarArray(2, width=0.1, num_distribution=_num_distribution),
        na.UniformUncertainScalarArray(
            na.ScalarLinearSpace(2, 3, axis='x', num=_num_x),
            width=0.1,
            num_distribution=_num_distribution,
        ),
    )
    units = (1, u.mm)
    arrays = tuple(
        na.UncertainScalarPoissionRandomSample(
            center=center * unit,
            shape_random=dict(y=_num_y),
        )
        for center in centers
        for unit in units
    )
    return arrays


@pytest.mark.parametrize("array", _uncertain_scalar_poisson_random_samples())
class TestUncertainScalarPoissonRandomSample(
    AbstractTestAbstractUncertainScalarRandomSample,
    named_arrays.tests.test_core.AbstractTestAbstractPoissonRandomSample,
):
    pass



class AbstractTestAbstractParameterizedUncertainScalarArray(
    AbstractTestAbstractImplicitUncertainScalarArray,
    named_arrays.tests.test_core.AbstractTestAbstractParameterizedArray,
):
    pass


class TestUncertainScalarLinearSpace(
    AbstractTestAbstractUncertainScalarRandomSample,
    named_arrays.tests.test_core.AbstractTestAbstractPoissonRandomSample,
):
    pass


class AbstractTestAbstractScalarSpace(
    AbstractTestAbstractParameterizedUncertainScalarArray,
    named_arrays.tests.test_core.AbstractTestAbstractSpace,
):
    pass


def _uncertain_scalar_linear_spaces() -> tuple[na.UncertainScalarLinearSpace, ...]:

    starts = (
        na.UniformUncertainScalarArray(
            na.ScalarLinearSpace(0, 1, axis='x', num=_num_x),
            width=0.2,
            num_distribution=_num_distribution,
        ),
    )
    stops = (
        2,
        na.ScalarLinearSpace(2, 3, axis='x', num=_num_x),
        na.UniformUncertainScalarArray(
            na.ScalarLinearSpace(2, 3, axis='x', num=_num_x),
            width=0.1,
            num_distribution=_num_distribution,
        ),
    )
    units = (1, u.mm)
    arrays = tuple(
        na.UncertainScalarLinearSpace(
            start=start * unit,
            stop=stop * unit,
            axis='y',
            num=_num_y,
        )
        for start in starts
        for stop in stops
        for unit in units
    )
    return arrays


@pytest.mark.parametrize("array", _uncertain_scalar_linear_spaces())
class TestUncertainScalarLinearSpace(
    AbstractTestAbstractScalarSpace,
    named_arrays.tests.test_core.AbstractTestAbstractLinearSpace,
):
    @pytest.mark.parametrize(
        argnames="axis",
        argvalues=[
            None,
            "x",
            "y",
            "z",
        ]
    )
    def test_volume_cell(
            self,
            array: na.ScalarLinearSpace,
            axis: None | str | Sequence[str],
    ):
        super().test_volume_cell(array=array, axis=axis)

        axis_ = na.axis_normalized(array, axis)
        if len(axis_) != 1:
            with pytest.raises(ValueError):
                array.volume_cell(axis)
            return

        if not set(axis_).issubset(array.shape):
            with pytest.raises(ValueError):
                array.volume_cell(axis)
            return

        assert np.allclose(array.volume_cell(axis), array.explicit.volume_cell(axis))



"""
A thin wrapper around the :mod:`colorsynth` package for named arrays.
"""

from __future__ import annotations
from typing import Callable
import named_arrays as na

__all__ = [
    "rgb",
    "colorbar",
    "rgb_and_colorbar",
]


def rgb(
    spd: na.AbstractArray,
    wavelength: None | na.AbstractScalar = None,
    axis: None | str = None,
    spd_min: None | na.ArrayLike = None,
    spd_max: None | na.ArrayLike = None,
    spd_norm: None | Callable = None,
    wavelength_min: None | na.ScalarLike = None,
    wavelength_max: None | na.ScalarLike = None,
    wavelength_norm: None | Callable = None,
) -> na.AbstractArray:
    """
    A thin wrapper around :func:`colorsynth.rgb()` for named arrays.

    Parameters
    ----------
    spd
        A spectral power distribution to convert to a RGB array.
    wavelength
        The wavelength coordinates corresponding to the spectral power distribution.
        If :obj:`None` (the default), the wavelength is assumed to be evenly
        sampled over the human visible range.
    axis
        The logical axis corresponding to changing wavelength.
        If :obj:`None` (the default), `spd` must be one-dimensional.
    spd_min
        The value of the spectral power distribution representing minimum
        intensity
    spd_max
        The vale of the spectral power distribution representing maximum
        intensity
    spd_norm
        An optional function that transforms the spectral power distribution
        values before mapping to RGB.
    wavelength_min
        The wavelength value that is mapped to the minimum wavelength of the
        human visible color range, 380 nm.
    wavelength_max
        The wavelength value that is mapped to the maximum wavelength of the
        human visible color range, 700 nm
    wavelength_norm
        An optional function to transform the wavelength values before they
        are mapped into the human visible color range.

    See Also
    --------
    :func:`colorsynth.rgb` :  Equivalent function for instances of :class:`numpy.ndarray`.

    Examples
    --------
    Plot the color of a random 3d cube

    .. jupyter-execute::

        import matplotlib.pyplot as plt
        import named_arrays as na

        # Define a random 3d cube
        a = na.random.uniform(
            low=0,
            high=1,
            shape_random=dict(x=16, y=16, wavelength=11)
        )

        # Compute the RGB colors of the 3d cube
        rgb = na.colorsynth.rgb(
            spd=a,
            axis="wavelength",
        )

        # Plot the RGB image
        fig, ax = plt.subplots(constrained_layout=True)
        na.plt.imshow(
            rgb,
            axis_x="x",
            axis_y="y",
            axis_rgb="wavelength",
            ax=ax,
        );
    """
    return na._named_array_function(
        func=rgb,
        spd=spd,
        wavelength=wavelength,
        axis=axis,
        spd_min=spd_min,
        spd_max=spd_max,
        spd_norm=spd_norm,
        wavelength_min=wavelength_min,
        wavelength_max=wavelength_max,
        wavelength_norm=wavelength_norm,
    )


def colorbar(
    spd: na.AbstractArray,
    wavelength: None | na.AbstractScalar = None,
    axis: None | str = None,
    spd_min: None | na.ArrayLike = None,
    spd_max: None | na.ArrayLike = None,
    spd_norm: None | Callable = None,
    wavelength_min: None | na.ScalarLike = None,
    wavelength_max: None | na.ScalarLike = None,
    wavelength_norm: None | Callable = None,
) -> na.FunctionArray[na.Cartesian2dVectorArray, na.AbstractScalar]:
    """
    A thin wrapper around :func:`colorsynth.colorbar()` for named arrays.

    Parameters
    ----------
    spd
        A spectral power distribution to convert to a RGB array.
    wavelength
        The wavelength coordinates corresponding to the spectral power distribution.
        If :obj:`None` (the default), the wavelength is assumed to be evenly
        sampled over the human visible range.
    axis
        The logical axis corresponding to changing wavelength.
        If :obj:`None` (the default), `spd` must be one-dimensional.
    spd_min
        The value of the spectral power distribution representing minimum
        intensity
    spd_max
        The vale of the spectral power distribution representing maximum
        intensity
    spd_norm
        An optional function that transforms the spectral power distribution
        values before mapping to RGB.
    wavelength_min
        The wavelength value that is mapped to the minimum wavelength of the
        human visible color range, 380 nm.
    wavelength_max
        The wavelength value that is mapped to the maximum wavelength of the
        human visible color range, 700 nm
    wavelength_norm
        An optional function to transform the wavelength values before they
        are mapped into the human visible color range.

    See Also
    --------
    :func:`colorsynth.colorbar` :  Equivalent function for instances of :class:`numpy.ndarray`.

    Examples
    --------

    Plot the colorbar of a random, 3D cube.

    .. jupyter-execute::

        import matplotlib.pyplot as plt
        import astropy.units as u
        import astropy.visualization
        import named_arrays as na

        # Define a random 3d cube
        a = na.random.uniform(
            low=0 * u.photon,
            high=1000 * u.photon,
            shape_random=dict(x=16, y=16, wavelength=11),
        )

        # Define wavelength axis
        wavelength = na.linspace(
            start=100 * u.AA,
            stop=200 * u.AA,
            axis="wavelength",
            num=a.shape["wavelength"],
        )

        # Compute the colorbar corresponding to the random 3d cube.
        colorbar = na.colorsynth.colorbar(
            spd=a,
            wavelength=wavelength,
            axis="wavelength",
        )

        # Plot the colorbar
        with astropy.visualization.quantity_support():
            fig, ax = plt.subplots()
            na.plt.pcolormesh(
                C=colorbar,
                axis_rgb="wavelength",
                ax=ax,
            )
    """
    return na._named_array_function(
        func=colorbar,
        spd=spd,
        wavelength=wavelength,
        axis=axis,
        spd_min=spd_min,
        spd_max=spd_max,
        spd_norm=spd_norm,
        wavelength_min=wavelength_min,
        wavelength_max=wavelength_max,
        wavelength_norm=wavelength_norm,
    )


def rgb_and_colorbar(
    spd: na.AbstractArray,
    wavelength: None | na.AbstractScalar = None,
    axis: None | str = None,
    spd_min: None | na.ArrayLike = None,
    spd_max: None | na.ArrayLike = None,
    spd_norm: None | Callable = None,
    wavelength_min: None | na.ScalarLike = None,
    wavelength_max: None | na.ScalarLike = None,
    wavelength_norm: None | Callable = None,
) -> tuple[
    na.AbstractArray,
    na.FunctionArray[na.Cartesian2dVectorArray, na.AbstractScalar],
]:
    """
    Convenience function that calls :func:`rgb` and :func:`colorbar` and
    returns the results as a tuple.

    Parameters
    ----------
    spd
        A spectral power distribution to convert to a RGB array.
    wavelength
        The wavelength coordinates corresponding to the spectral power distribution.
        If :obj:`None` (the default), the wavelength is assumed to be evenly
        sampled over the human visible range.
    axis
        The logical axis corresponding to changing wavelength.
        If :obj:`None` (the default), `spd` must be one-dimensional.
    spd_min
        The value of the spectral power distribution representing minimum
        intensity
    spd_max
        The vale of the spectral power distribution representing maximum
        intensity
    spd_norm
        An optional function that transforms the spectral power distribution
        values before mapping to RGB.
    wavelength_min
        The wavelength value that is mapped to the minimum wavelength of the
        human visible color range, 380 nm.
    wavelength_max
        The wavelength value that is mapped to the maximum wavelength of the
        human visible color range, 700 nm
    wavelength_norm
        An optional function to transform the wavelength values before they
        are mapped into the human visible color range.

    See Also
    --------
    :func:`colorsynth.colorbar` :  Equivalent function for instances of :class:`numpy.ndarray`.

    Examples
    --------

    Plot the colorbar of a random, 3D cube.

    .. jupyter-execute::

        import matplotlib.pyplot as plt
        import astropy.units as u
        import astropy.visualization
        import named_arrays as na

        # Define a random 3d cube
        a = na.random.uniform(
            low=0 * u.photon,
            high=1000 * u.photon,
            shape_random=dict(x=16, y=16, wavelength=11),
        )

        # Define wavelength axis
        wavelength = na.linspace(
            start=100 * u.AA,
            stop=200 * u.AA,
            axis="wavelength",
            num=a.shape["wavelength"],
        )

        # Compute the colorbar corresponding to the random 3d cube.
        rgb, colorbar = na.colorsynth.rgb_and_colorbar(
            spd=a,
            wavelength=wavelength,
            axis="wavelength",
        )

        # Plot the colorbar
        with astropy.visualization.quantity_support():
            fig, axs = plt.subplots(
                ncols=2,
                gridspec_kw=dict(width_ratios=[.9,.1]),
                constrained_layout=True,
            )
            na.plt.imshow(
                rgb,
                axis_x="x",
                axis_y="y",
                axis_rgb="wavelength",
                ax=axs[0],
            );
            na.plt.pcolormesh(
                C=colorbar,
                axis_rgb="wavelength",
                ax=axs[1],
            )
            axs[1].yaxis.tick_right()
            axs[1].yaxis.set_label_position("right")

    """
    kwargs = dict(
        spd=spd,
        wavelength=wavelength,
        axis=axis,
        spd_min=spd_min,
        spd_max=spd_max,
        spd_norm=spd_norm,
        wavelength_min=wavelength_min,
        wavelength_max=wavelength_max,
        wavelength_norm=wavelength_norm,
    )
    result_rgb = rgb(**kwargs)
    result_colorbar = colorbar(**kwargs)
    return result_rgb, result_colorbar



from typing import Sequence, Literal, Callable
import pytest
import numpy as np
import astropy.units as u
import named_arrays as na
import named_arrays.tests.test_core
import named_arrays._vectors.cartesian.tests.test_vectors_cartesian_2d
from . import test_functions

__all__ = [
    "AbstractTestAbstractFunctionArrayVertices",
]

_num_x = named_arrays.tests.test_core.num_x
_num_y = named_arrays.tests.test_core.num_y
_num_z = named_arrays.tests.test_core.num_z
_num_distribution = named_arrays.tests.test_core.num_distribution


def _function_arrays():
    functions_1d = [
        na.FunctionArray(
            inputs=na.ScalarLinearSpace(0, 1, axis='y', num=_num_y+1),
            outputs=na.ScalarUniformRandomSample(-5, 5, shape_random=dict(y=_num_y))
        )
    ]

    inputs_2d = [
        na.ScalarLinearSpace(0, 1, axis='y', num=_num_y+1),
        na.Cartesian2dVectorLinearSpace(
            start=0,
            stop=1,
            axis=na.Cartesian2dVectorArray('x', 'y'),
            num=na.Cartesian2dVectorArray(_num_x+1, _num_y+1)
        )
    ]

    outputs_2d = [
        na.ScalarUniformRandomSample(-5, 5, shape_random=dict(x=_num_x, y=_num_y, z=_num_z)),
    ]

    functions_2d = [
        na.FunctionArray(
            inputs=inputs,
            outputs=outputs,
        )
        for inputs in inputs_2d
        for outputs in outputs_2d
    ]

    functions = functions_1d + functions_2d

    return functions


def _function_arrays_2():
    return (
        6,
        na.ScalarArray(6),
        na.UniformUncertainScalarArray(6, width=1, num_distribution=_num_distribution),
        na.Cartesian2dVectorArray(6, 7),
        na.FunctionArray(
            inputs=na.ScalarLinearSpace(0, 1, axis='y', num=_num_y+1),
            outputs=na.ScalarUniformRandomSample(-6, 6, shape_random=dict(y=_num_y))
        ),
        na.FunctionArray(
            inputs=na.ScalarLinearSpace(0, 1, axis='y', num=_num_y+1),
            outputs=na.Cartesian2dVectorUniformRandomSample(-6, 6, shape_random=dict(y=_num_y))
        )
    )


class AbstractTestAbstractFunctionArrayVertices(
    test_functions.AbstractTestAbstractFunctionArray,
):

    @pytest.mark.parametrize('newshape', [dict(r=-1)])
    def test_reshape(self, array: na.AbstractArray, newshape: dict[str, int]):

        for ax in newshape:
            if ax in array.axes_vertex or (ax not in array.axes and len(array.axes_vertex) != 0):
                with pytest.raises(ValueError):
                    result = np.reshape(a=array, newshape=newshape)
                return

        super().test_reshape(array, newshape)
    @pytest.mark.parametrize('axes', [None, ('x', 'y'), ('x', 'y', 'z')])
    def test_combine_axes(
            self,
            array: na.AbstractArray,
            axes: None | Sequence[str]
    ):
        axis_new = 'new_test_axis'
        if axes is None and len(array.axes_vertex) != 0:
            with pytest.raises(ValueError):
                array.combine_axes(axes=axes, axis_new=axis_new)
            return

        if np.any([ax in array.axes_vertex for ax in axes]):
            with pytest.raises(ValueError):
                array.combine_axes(axes=axes, axis_new=axis_new)
            return

        super().test_combine_axes(array=array, axes=axes)


    @pytest.mark.parametrize(
        argnames='item',
        argvalues=[
            dict(y=0),
            dict(y=slice(0, 1)),
        ]

    )
    def test__getitem__(
            self,
            array: na.AbstractFunctionArray,
            item: dict[str, int | slice | na.AbstractArray] | na.AbstractArray
    ):

        super().test__getitem__(array=array, item=item)


    @pytest.mark.parametrize("array_2", _function_arrays_2())
    class TestUfuncBinary(
        test_functions.AbstractTestAbstractFunctionArray.TestUfuncBinary
    ):
        pass

    @pytest.mark.parametrize("array_2", _function_arrays_2())
    class TestMatmul(
        test_functions.AbstractTestAbstractFunctionArray.TestMatmul
    ):
        pass

    class TestArrayFunctions(
        test_functions.AbstractTestAbstractFunctionArray.TestArrayFunctions
    ):
        @pytest.mark.parametrize('newshape', [dict(r=-1)])
        def test_reshape(self, array: na.AbstractArray, newshape: dict[str, int]):

            for ax in newshape:
                if ax in array.axes_vertex or (ax not in array.axes and len(array.axes_vertex) != 0):
                    with pytest.raises(ValueError):
                        result = np.reshape(a=array, newshape=newshape)
                    return

            super().test_reshape(array, newshape)

        @pytest.mark.parametrize('axis', ['x', 'y'])
        def test_concatenate(
                self,
                array: na.AbstractArray,
                axis: str,
        ):
            arrays = [array,array]
            if axis in array.axes_vertex:
                with pytest.raises(ValueError):
                    np.concatenate(arrays, axis=axis)
                return
            super().test_concatenate(array, axis)

        def test_nonzero(self, array: na.AbstractArray):
            if len(array.axes_vertex) != 0:
                with pytest.raises(ValueError, match=f"item not supported by array with type {type(array)}"):
                    #matches test_core, but is a bad test. nonzero only applied to boolean arrays
                    array[np.nonzero(array>array.mean())]
                return
            super().test_nonzero(array=array)

        class TestReductionFunctions(
            test_functions.AbstractTestAbstractFunctionArray.TestArrayFunctions.TestReductionFunctions
        ):
            @pytest.mark.parametrize(
                argnames='where',
                argvalues=[
                    np._NoValue,
                    True,
                    na.ScalarArray(True),
                    na.FunctionArray(
                        inputs=na.Cartesian2dVectorLinearSpace(
                            start=0,
                            stop=1,
                            axis=na.Cartesian2dVectorArray('x', 'y'),
                            num=na.Cartesian2dVectorArray(_num_x + 1, _num_y + 1)
                        ),
                        outputs=(na.ScalarLinearSpace(-1, 1, 'x', _num_x) >= 0)
                                | (na.ScalarLinearSpace(-1, 1, 'y', _num_y) >= 0)
                    )
                ]
            )
            def test_reduction_functions(
                    self,
                    func: Callable,
                    array: na.AbstractFunctionArray,
                    axis: None | str | Sequence[str],
                    dtype: None | type | np.dtype,
                    keepdims: bool,
                    where: bool | na.AbstractFunctionArray,
            ):
                super().test_reduction_functions(
                    func=func,
                    array=array,
                    axis=axis,
                    dtype=dtype,
                    keepdims=keepdims,
                    where=where,
                )

    class TestNamedArrayFunctions(
        test_functions.AbstractTestAbstractFunctionArray.TestNamedArrayFunctions
    ):
        class TestHistogram(
            test_functions.AbstractTestAbstractFunctionArray.TestNamedArrayFunctions.TestHistogram,
        ):
            def test_histogram(
                self,
                array: na.AbstractFunctionArray,
                bins: Literal["dict"],
                axis: None | str | Sequence[str],
                min: None | na.AbstractScalarArray | na.AbstractVectorArray,
                max: None | na.AbstractScalarArray | na.AbstractVectorArray,
                weights: None | na.AbstractScalarArray,
            ):

                axis_normalized = tuple(array.shape) if axis is None else (axis,) if isinstance(axis, str) else axis
                for ax in axis_normalized:
                    if ax in array.axes_vertex:
                        with pytest.raises(ValueError, match="Taking a histogram of a histogram doesn't work right now."):
                            na.histogram(
                                a=array,
                                bins=bins,
                                axis=axis,
                                min=min,
                                max=max,
                                weights=weights,
                            )
                        return

                super().test_histogram(
                    array=array,
                    bins=bins,
                    axis=axis,
                    min=min,
                    max=max,
                    weights=weights,
                )


@pytest.mark.parametrize("array", _function_arrays())
class TestFunctionArrayVertices(
    AbstractTestAbstractFunctionArrayVertices,
    named_arrays.tests.test_core.AbstractTestAbstractExplicitArray,
):

    @pytest.mark.parametrize(
        argnames="item",
        argvalues=[
            dict(y=0),
            dict(x=0, y=0),
            dict(y=slice(None)),
            # dict(y=na.ScalarArrayRange(0, _num_y, axis='y')),
            # dict(x=na.ScalarArrayRange(0, _num_x, axis='x'), y=na.ScalarArrayRange(0, _num_y, axis='y')),
            na.FunctionArray(
                inputs=na.ScalarLinearSpace(0, 1, axis='y', num=_num_y+1),
                outputs=na.ScalarArray.ones(shape=dict(y=_num_y), dtype=bool),
            )
        ]
    )
    @pytest.mark.parametrize(
        argnames="value",
        argvalues=[
            0,
            na.FunctionArray(
                inputs=na.ScalarLinearSpace(0, 1, axis='y', num=_num_y+1),
                outputs=na.ScalarUniformRandomSample(-5, 5, dict(y=_num_y)),
            )
        ]
    )
    def test__setitem__(
            self,
            array: na.AbstractArray,
            item: dict[str, int | slice | na.ScalarArray] | na.AbstractFunctionArray,
            value: float | u.Quantity | na.AbstractScalar | na.AbstractVectorArray,
    ):
        super().test__setitem__(array=array.explicit, item=item, value=value)


@pytest.mark.parametrize("type_array", [na.FunctionArray])
class TestFunctionArrayCreation(
    named_arrays.tests.test_core.AbstractTestAbstractExplicitArrayCreation,
):
    @pytest.mark.parametrize("like", [None] + _function_arrays())
    class TestFromScalarArray(
        named_arrays.tests.test_core.AbstractTestAbstractExplicitArrayCreation.TestFromScalarArray,
    ):
        pass


"""
Array resampling and interpolation.

A wrapper around the :mod:`regridding` module for named arrays.
"""

from __future__ import annotations
from typing import Sequence, Literal
import astropy.units as u
import named_arrays as na

__all__ = [
    "regrid",
    "weights",
    "regrid_from_weights",
]


def regrid(
    coordinates_input: na.AbstractScalar | na.AbstractVectorArray,
    coordinates_output: float | u.Quantity,
    values_input: na.AbstractScalarArray,
    axis_input: None | Sequence[str] = None,
    axis_output: None | Sequence[str] = None,
    method: Literal['multilinear', 'conservative'] = 'multilinear',
) -> na.AbstractScalarArray:
    """
    Regrid an array of values defined on a logically-rectangular curvilinear
    grid onto a new logically-rectangular curvilinear grid.

    Parameters
    ----------
    coordinates_input
        Coordinates of the input grid.
    coordinates_output
        Coordinates of the output grid.
        Should have the same number of components as the input grid.
    values_input
        Input array of values to be resampled.
    axis_input
        Logical axes of the input grid to resample.
        If :obj:`None`, resample all the axes of the input grid.
        The number of axes should be equal to the number of
        coordinates in the input grid.
    axis_output
        Logical axes of the output grid corresponding to the resampled axes
        of the input grid.
        If :obj:`None`, all the axes of the output grid correspond to resampled
        axes in the input grid.
        The number of axes should be equal to the number of
        coordinates in the output grid.
    method
        The type of regridding to use.

    Examples
    --------

    Regrid a 2D array using conservative resampling.

    .. jupyter-execute::

        import numpy as np
        import matplotlib.pyplot as plt
        import named_arrays as na

        # Define the number of edges in the input grid
        num_x = 66
        num_y = 66

        # Define a dummy linear grid
        x = na.linspace(-5, 5, axis="x", num=num_x)
        y = na.linspace(-5, 5, axis="y", num=num_y)

        # Define the curvilinear input grid using the dummy grid
        angle = 0.4
        coordinates_input = na.Cartesian2dVectorArray(
            x=x * np.cos(angle) - y * np.sin(angle) + 0.05 * x * x,
            y=x * np.sin(angle) + y * np.cos(angle) + 0.05 * y * y,
        )

        # Define the test pattern
        a_input = np.cos(np.square(x)) * np.cos(np.square(y))
        a_input = a_input.cell_centers()

        # Define a rectilinear output grid using the limits of the input grid
        coordinates_output = na.Cartesian2dVectorLinearSpace(
            start=coordinates_input.min(),
            stop=coordinates_input.max(),
            axis=na.Cartesian2dVectorArray("x2", "y2"),
            num=66,
        )

        # Regrid the test pattern onto the new grid
        a_output = na.regridding.regrid(
            coordinates_input=coordinates_input,
            coordinates_output=coordinates_output,
            values_input=a_input,
            method="conservative",
        )

        fig, ax = plt.subplots(
            ncols=2,
            sharex=True,
            sharey=True,
            figsize=(8, 4),
            constrained_layout=True,
        );
        na.plt.pcolormesh(coordinates_input, C=a_input, ax=ax[0])
        na.plt.pcolormesh(coordinates_output, C=a_output, ax=ax[1])
        ax[0].set_title("input array");
        ax[1].set_title("regridded array");
    """

    _weights, shape_input, shape_output = weights(
        coordinates_input=coordinates_input,
        coordinates_output=coordinates_output,
        axis_input=axis_input,
        axis_output=axis_output,
        method=method
    )

    result = regrid_from_weights(
        weights=_weights,
        shape_input=shape_input,
        shape_output=shape_output,
        values_input=values_input,
        # values_output=values_output,
        # axis_input=axis_input,
        # axis_output=axis_output,
    )

    return result


def weights(
    coordinates_input: na.AbstractScalar | na.AbstractVectorArray,
    coordinates_output: na.AbstractScalar | na.AbstractVectorArray,
    axis_input: None | str | Sequence[str] = None,
    axis_output: None | str | Sequence[str] = None,
    method: Literal['multilinear', 'conservative'] = 'multilinear',
) -> tuple[na.AbstractScalar, dict[str, int], dict[str, int]]:
    """
    Save the results of a regridding operation as a sequence of weights,
    which can be used in subsequent regridding operations on the same grid.

    The results of this function are designed to be used by
    :func:`regrid_from_weights`

    This function returns a tuple containing a ragged array of weights,
    the shape of the input coordinates, and the shape of the output coordinates.

    Parameters
    ----------
    coordinates_input
        Coordinates of the input grid.
    coordinates_output
        Coordinates of the output grid.
        Should have the same number of coordinates as the input grid.
    axis_input
        Logical axes of the input grid to resample.
        If :obj:`None`, resample all the axes of the input grid.
        The number of axes should be equal to the number of
        coordinates in the input grid.
    axis_output
        Logical axes of the output grid corresponding to the resampled axes
        of the input grid.
        If :obj:`None`, all the axes of the output grid correspond to resampled
        axes in the input grid.
        The number of axes should be equal to the number of
        coordinates in the output grid.
    method
        The type of regridding to use.

    See Also
    --------
    :func:`regridding.weights`: An equivalent function for instances of :class:`numpy.ndarray`.
    :func:`regrid_from_weights`: A function designed to use the outputs of this function.
    :func:`regrid`: Resample an array without saving the weights

    """
    return na._named_array_function(
        func=weights,
        coordinates_input=coordinates_input,
        coordinates_output=coordinates_output,
        axis_input=axis_input,
        axis_output=axis_output,
        method=method,
    )


def regrid_from_weights(
    weights: na.AbstractScalar,
    shape_input: dict[str, int],
    shape_output: dict[str, int],
    values_input: na.AbstractScalar | na.AbstractVectorArray,
) -> na.AbstractArray:
    """
    Regrid an array of values using weights computed by
    :func:`weights`.

    Parameters
    ----------
    weights
        Ragged array of weights computed by :func:`weights`.
    shape_input
        Broadcasted shape of the input coordinates computed by :func:`weights`.
    shape_output
        Broadcasted shape of the output coordinates computed by :func:`weights`.
    values_input
        Input array of values to be resampled.
    """
    return na._named_array_function(
        func=regrid_from_weights,
        weights=weights,
        shape_input=shape_input,
        shape_output=shape_output,
        values_input=values_input,
    )



from ._named_array_functions import *
from . import random
from . import plt
from . import optimize
from . import regridding
from . import transformations
from . import ndfilters
from . import colorsynth
from . import regridding
from ._core import *
from ._scalars.scalars import *
from ._scalars.uncertainties.uncertainties import *
from ._vectors.vectors import *
from ._vectors.cartesian.vectors_cartesian import *
from ._vectors.cartesian.vectors_cartesian_2d import *
from ._vectors.cartesian.vectors_cartesian_3d import *
from ._vectors.cartesian.vectors_cartesian_nd import *
from ._vectors.vectors_temporal import *
from ._vectors.vectors_spectral import *
from ._vectors.vectors_positional import *
from ._vectors.vectors_directional import *
from ._vectors.vectors_spectral_positional import *
from ._vectors.vectors_spectral_directional import *
from ._vectors.vectors_temporal_spectral import *
from ._vectors.vectors_temporal_spectral_positional import *
from ._vectors.vectors_temporal_spectral_directional import *
from ._vectors.vectors_input_output import *
from ._matrices.matrices import *
from ._matrices.cartesian.matrices_cartesian import *
from ._matrices.cartesian.matrices_cartesian_2d import *
from ._matrices.cartesian.matrices_cartesian_3d import *
from ._matrices.cartesian.matrices_cartesian_nd import *
from ._matrices.matrices_temporal import *
from ._matrices.matrices_spectral import *
from ._matrices.matrices_positional import *
from ._matrices.matrices_directional import *
from ._matrices.matrices_spectral_positional import *
from ._matrices.matrices_spectral_directional import *
from ._matrices.matrices_temporal_spectral import *
from ._matrices.matrices_temporal_spectral_positional import *
from ._matrices.matrices_temporal_spectral_directional import *
from ._matrices.matrices_input_output import *
from ._functions.functions import *



from typing import Callable, Sequence, Any, Literal
import collections
import dataclasses
import numpy as np
import numpy.typing as npt
import matplotlib.axes
import matplotlib.artist
import matplotlib.pyplot as plt
import matplotlib.animation
import astropy.units as u
import astroscrappy
import ndfilters
import colorsynth
import regridding
import named_arrays as na
from . import scalars

__all__ = [
    "ASARRAY_LIKE_FUNCTIONS",
    "RANDOM_FUNCTIONS",
    "PLT_PLOT_LIKE_FUNCTIONS",
    "PLT_AXES_SETTERS",
    "PLT_AXES_GETTERS",
    "PLT_AXES_ATTRIBUTES",
    "HANDLED_FUNCTIONS",
    "random",
    "jacobian",
    "colorsynth_rgb",
]

ASARRAY_LIKE_FUNCTIONS = (
    na.asarray,
    na.asanyarray,
)
RANDOM_FUNCTIONS = (
    na.random.uniform,
    na.random.normal,
    na.random.poisson,
)
PLT_PLOT_LIKE_FUNCTIONS = (
    na.plt.plot,
    na.plt.fill,
)
PLT_AXES_SETTERS = (
    na.plt.set_xlabel,
    na.plt.set_ylabel,
    na.plt.set_title,
    na.plt.set_xscale,
    na.plt.set_yscale,
    na.plt.set_aspect,
)
PLT_AXES_GETTERS = (
    na.plt.get_xlabel,
    na.plt.get_ylabel,
    na.plt.get_title,
    na.plt.get_xscale,
    na.plt.get_yscale,
    na.plt.get_aspect,
    na.plt.twinx,
    na.plt.twiny,
    na.plt.invert_xaxis,
    na.plt.invert_yaxis,
)
PLT_AXES_ATTRIBUTES = (
    na.plt.transAxes,
    na.plt.transData,
)
NDFILTER_FUNCTIONS = (
    na.ndfilters.mean_filter,
    na.ndfilters.trimmed_mean_filter,
    na.ndfilters.variance_filter,
)
HANDLED_FUNCTIONS = dict()


def _implements(function: Callable):
    """Register a __named_array_function__ implementation for AbstractScalarArray objects."""
    def decorator(func):
        HANDLED_FUNCTIONS[function] = func
        return func
    return decorator


def asarray_like(
        func: Callable,
        a: None | float | u.Quantity | na.AbstractScalarArray,
        dtype: None | type | np.dtype = None,
        order: None | str = None,
        *,
        like: None | na.AbstractScalarArray = None,
) -> None | na.ScalarArray:

    func_numpy = getattr(np, func.__name__)

    if isinstance(a, na.AbstractArray):
        if isinstance(a, na.AbstractScalarArray):
            a_ndarray = a.ndarray
            a_axes = a.axes
        else:
            return NotImplemented
    else:
        a_ndarray = a
        a_axes = None

    if isinstance(like, na.AbstractArray):
        if isinstance(like, na.AbstractScalarArray):
            like_ndarray = like.ndarray
            type_like = like.type_explicit
        else:
            return NotImplemented
    else:
        like_ndarray = like
        type_like = na.ScalarArray

    if isinstance(like_ndarray, u.Quantity):
        like_ndarray = like_ndarray.value
    like_ndarray = func_numpy(like_ndarray)

    return type_like(
        ndarray=func_numpy(
            a=a_ndarray,
            dtype=dtype,
            order=order,
            like=like_ndarray,
        ),
        axes=a_axes,
    )


@_implements(na.arange)
def arange(
        start: float | complex | u.Quantity | na.AbstractArray,
        stop: float | complex | u.Quantity | na.AbstractArray,
        axis: str | na.AbstractArray,
        step: int | na.AbstractArray = 1,
) -> na.ScalarArray:

    start = scalars._normalize(start)
    stop = scalars._normalize(stop)

    if start.size > 1:
        raise ValueError(f"`start` must have only one element, got shape {start.shape}")

    if stop.size > 1:
        raise ValueError(f"`stop` must have only one element, got shape {stop.shape}")

    return na.ScalarArray(
        ndarray=np.arange(
            start=start.ndarray,
            stop=stop.ndarray,
            step=step,
        ),
        axes=(axis,),
    )


@_implements(na.unit)
def unit(
        a: na.AbstractScalarArray,
        unit_dimensionless: None | float | u.UnitBase = None,
        squeeze: bool = True,
) -> None | u.UnitBase:
    return na.unit(
        a=a.ndarray,
        unit_dimensionless=unit_dimensionless,
    )


@_implements(na.unit_normalized)
def unit_normalized(
        a: na.AbstractScalarArray,
        unit_dimensionless: float | u.UnitBase = u.dimensionless_unscaled,
        squeeze: bool = True,
) -> u.UnitBase:
    result = na.unit(
        a=a,
        unit_dimensionless=unit_dimensionless,
    )
    return result


@_implements(na.interp)
def interp(
        x: float | u.Quantity | na.AbstractScalarArray,
        xp:  na.AbstractScalarArray,
        fp: na.AbstractScalarArray,
        axis: None | str = None,
        left: None | float | u.Quantity | na.AbstractScalarArray = None,
        right: None | float | u.Quantity | na.AbstractScalarArray = None,
        period: None | float | u.Quantity | na.AbstractScalarArray = None,
):
    try:
        x = scalars._normalize(x)
        xp = scalars._normalize(xp)
        fp = scalars._normalize(fp)
        left = scalars._normalize(left) if left is not None else left
        right = scalars._normalize(right) if right is not None else right
        period = scalars._normalize(period) if period is not None else period
    except na.ScalarTypeError:
        return NotImplemented

    if axis is None:
        if xp.ndim != 1:
            raise ValueError("if `axis` is `None`, `xp` must have only one axis")
        axis = next(iter(xp.shape))

    shape = na.shape_broadcasted(
        # x,
        xp[{axis: 0}],
        fp[{axis: 0}],
        left,
        right,
        period,
    )

    x = na.broadcast_to(x, na.broadcast_shapes(x.shape, shape))
    xp = na.broadcast_to(xp, na.broadcast_shapes(xp.shape, shape))
    fp = na.broadcast_to(fp, na.broadcast_shapes(fp.shape, shape))
    left = na.broadcast_to(left, shape) if left is not None else left
    right = na.broadcast_to(right, shape) if right is not None else right
    period = na.broadcast_to(period, shape) if period is not None else period

    result = np.empty_like(x.value, dtype=fp.dtype)
    if fp.unit is not None:
        result = result << fp.unit

    for index in na.ndindex(shape):
        x_index = x[index]
        xp_index = xp[index]
        fp_index = fp[index]
        left_index = left[index].ndarray if left is not None else left
        right_index = right[index].ndarray if right is not None else right
        period_index = period[index].ndarray if period is not None else period

        shape_index = na.shape_broadcasted(xp_index, fp_index)

        result[index] = result.type_explicit(
            ndarray=np.interp(
                x=x_index.ndarray,
                xp=xp_index.ndarray_aligned(shape_index),
                fp=fp_index.ndarray_aligned(shape_index),
                left=left_index,
                right=right_index,
                period=period_index,
            ),
            axes=x_index.axes,
        )

    return result


@_implements(na.histogram)
def histogram(
    a: na.AbstractScalarArray,
    bins: dict[str, int] | na.AbstractScalarArray,
    axis: None | str | Sequence[str] = None,
    min: None | na.AbstractScalarArray = None,
    max: None | na.AbstractScalarArray = None,
    density: bool = False,
    weights: None | na.AbstractScalarArray = None,
) -> na.FunctionArray[na.AbstractScalarArray, na.ScalarArray]:

    if isinstance(bins, na.AbstractArray):
        bins = (bins, )

    hist, edges = na.histogramdd(
        a,
        bins=bins,
        axis=axis,
        min=min,
        max=max,
        density=density,
        weights=weights,
    )
    edges = edges[0]

    return na.FunctionArray(
        inputs=edges,
        outputs=hist,
    )


@_implements(na.histogram2d)
def histogram2d(
    x: na.AbstractScalarArray,
    y: na.AbstractScalarArray,
    bins: dict[str, int] | na.AbstractCartesian2dVectorArray,
    axis: None | str | Sequence[str] = None,
    min: None | na.AbstractScalarArray | na.AbstractCartesian2dVectorArray = None,
    max: None | na.AbstractScalarArray | na.AbstractCartesian2dVectorArray = None,
    density: bool = False,
    weights: None | na.AbstractScalarArray = None,
) -> na.FunctionArray[na.Cartesian2dVectorArray, na.ScalarArray]:
    try:
        x = scalars._normalize(x)
        y = scalars._normalize(y)
        weights = scalars._normalize(weights) if weights is not None else weights
    except scalars.ScalarTypeError:  # pragma: nocover
        return NotImplemented

    shape = na.shape_broadcasted(x, y, weights)

    x = na.broadcast_to(x, shape)
    y = na.broadcast_to(y, shape)
    weights = na.broadcast_to(weights, shape) if weights is not None else weights

    if axis is None:
        axis = tuple(shape)
    elif isinstance(axis, str):
        axis = (axis, )

    shape_orthogonal = {a: shape[a] for a in shape if a not in axis}

    if isinstance(bins, na.AbstractCartesian2dVectorArray):

        axis_x = set(bins.x.shape) - set(shape_orthogonal)
        if len(axis_x) != 1:  # pragma: nocover
            raise ValueError(
                f"if `bins` is a vector, `bins.x` must have exactly one new axis, "
                f"got {axis_x}"
            )

        axis_y = set(bins.y.shape) - set(shape_orthogonal)
        if len(axis_y) != 1:  # pragma: nocover
            raise ValueError(
                f"if `bins` is a vector, `bins.y` must have exactly one new axis, "
                f"got {axis_y}"
            )

        if axis_x == axis_y:  # pragma: nocover
            raise ValueError(
                f"if `bins` is a vector, `bins.x` and `bins.y` must be separable "
                f"along the new axes, found non-separable axis {axis_x}."
            )

        edges = bins

    elif isinstance(bins, dict):

        a = na.Cartesian2dVectorArray(x, y)

        if min is None:
            min = a.min(axis)
        elif not isinstance(min, na.AbstractCartesian2dVectorArray):
            min = na.Cartesian2dVectorArray(min, min)

        if max is None:
            max = a.max(axis)
        elif not isinstance(max, na.AbstractCartesian2dVectorArray):
            max = na.Cartesian2dVectorArray(max, max)

        if len(bins) != 2:  # pragma: nocover
            raise ValueError(
                f"if `bins` is a dictionary, it must have exactly two keys, "
                f"got {bins=}."
            )

        axis_x, axis_y = tuple(bins)

        edges = na.Cartesian2dVectorLinearSpace(
            start=min,
            stop=max,
            axis=na.Cartesian2dVectorArray(
                x=axis_x,
                y=axis_y,
            ),
            num=na.Cartesian2dVectorArray(
                x=bins[axis_x] + 1,
                y=bins[axis_y] + 1,
            ),
        )

    else:  # pragma: nocover
        return NotImplemented

    shape_edges_x = na.broadcast_shapes(shape_orthogonal, edges.x.shape)
    shape_edges_y = na.broadcast_shapes(shape_orthogonal, edges.y.shape)

    edges_broadcasted = dataclasses.replace(
        edges.explicit,
        x=edges.x.broadcast_to(shape_edges_x),
        y=edges.y.broadcast_to(shape_edges_y),
    )

    shape_edges = edges.shape
    shape_hist = {
        ax: shape_edges[ax] - 1
        for ax in shape_edges
        if ax not in shape_orthogonal
    }
    shape_hist = na.broadcast_shapes(shape_orthogonal, shape_hist)

    hist = na.ScalarArray.empty(shape_hist)

    unit_weights = na.unit(weights)
    hist = hist if unit_weights is None else hist << unit_weights

    for i in na.ndindex(shape_orthogonal):
        edges_x_i = edges_broadcasted.x[i]
        edges_y_i = edges_broadcasted.y[i]
        hist_i, _, _ = np.histogram2d(
            x=x[i].ndarray_aligned(axis).reshape(-1),
            y=y[i].ndarray_aligned(axis).reshape(-1),
            bins=(
                edges_x_i.ndarray,
                edges_y_i.ndarray,
            ),
            density=density,
            weights=weights[i].ndarray_aligned(axis).reshape(-1) if weights is not None else weights,
        )

        hist[i] = na.ScalarArray(
            ndarray=hist_i,
            axes=edges_x_i.axes + edges_y_i.axes,
        )

    return na.FunctionArray(
        inputs=edges,
        outputs=hist,
    )


@_implements(na.histogramdd)
def histogramdd(
    *sample: na.AbstractScalarArray,
    bins: dict[str, int] | na.AbstractScalarArray | Sequence[na.AbstractScalarArray],
    axis: None | str | Sequence[str] = None,
    min: None | na.AbstractScalarArray | Sequence[na.AbstractScalarArray] = None,
    max: None | na.AbstractScalarArray | Sequence[na.AbstractScalarArray] = None,
    density: bool = False,
    weights: None | na.AbstractScalarArray = None,
) -> tuple[na.AbstractScalarArray, tuple[na.AbstractScalarArray, ...]]:

    try:
        sample = [scalars._normalize(s) for s in sample]
        weights = scalars._normalize(weights) if weights is not None else weights
    except scalars.ScalarTypeError:  # pragma: nocover
        return NotImplemented

    shape = na.shape_broadcasted(*sample, weights)

    sample = [s.broadcast_to(shape) for s in sample]
    weights = na.broadcast_to(weights, shape) if weights is not None else weights

    if axis is None:
        axis = tuple(shape)
    elif isinstance(axis, str):
        axis = (axis,)

    if isinstance(bins, dict):

        if min is None:
            min = [s.min(axis) for s in sample]
        elif not isinstance(min, collections.abc.Sequence):
            min = [min] * len(sample)

        if max is None:
            max = [s.max(axis) for s in sample]
        elif not isinstance(max, collections.abc.Sequence):
            max = [max] * len(sample)

        try:
            max = [scalars._normalize(m) for m in max]
            min = [scalars._normalize(m) for m in min]
        except scalars.ScalarTypeError:  # pragma: nocover
            return NotImplemented

        if len(bins) != len(sample):  # pragma: nocover
            raise ValueError(
                f"if {bins=} is a dictionary, it must have the same number of"
                f"elements as {len(sample)=}."
            )

        bins = [
            na.linspace(start, stop, axis=ax, num=bins[ax])
            for start, stop, ax in zip(min, max, bins)
        ]

    try:
        bins = [scalars._normalize(b) for b in bins]
    except scalars.ScalarTypeError:  # pragma: nocover
        return NotImplemented

    shape_orthogonal = {a: shape[a] for a in shape if a not in axis}

    bins_broadcasted = [
        b.broadcast_to(na.broadcast_shapes(shape_orthogonal, b.shape))
        for b in bins
    ]

    shape_bins = na.shape_broadcasted(*bins)
    shape_hist = {
        ax: shape_bins[ax] - 1
        for ax in shape_bins
        if ax not in shape_orthogonal
    }
    shape_hist = na.broadcast_shapes(shape_orthogonal, shape_hist)

    hist = na.ScalarArray.empty(shape_hist)

    unit_weights = na.unit(weights)
    hist = hist if unit_weights is None else hist << unit_weights

    for i in na.ndindex(shape_orthogonal):
        hist_i, _ = np.histogramdd(
            sample=[s[i].ndarray_aligned(axis).reshape(-1) for s in sample],
            bins=[b[i].ndarray for b in bins_broadcasted],
            density=density,
            weights=weights[i].ndarray_aligned(axis).reshape(-1) if weights is not None else weights,
        )

        hist[i] = na.ScalarArray(
            ndarray=hist_i,
            axes=sum((b[i].axes for b in bins_broadcasted), ()),
        )

    return hist, tuple(bins)


def random(
        func: Callable,
        *args: float | u.Quantity | na.AbstractScalarArray,
        shape_random: None | dict[str, int] = None,
        seed: None | int = None,
        **kwargs: float | u.Quantity | na.AbstractScalarArray,
) -> na.ScalarArray:

    try:
        args = tuple(scalars._normalize(arg) for arg in args)
        kwargs = {k: scalars._normalize(kwargs[k]) for k in kwargs}
    except na.ScalarTypeError:
        return NotImplemented

    if shape_random is None:
        shape_random = dict()

    shape_base = na.shape_broadcasted(*args, *kwargs.values())
    shape = na.broadcast_shapes(shape_base, shape_random)

    args = tuple(arg.ndarray_aligned(shape) for arg in args)
    kwargs = {k: kwargs[k].ndarray_aligned(shape) for k in kwargs}

    unit = None
    for a in args + tuple(kwargs.values()):
        if isinstance(a, u.Quantity):
            unit = a.unit
            break

    if unit is not None:
        args = tuple(
            arg.to_value(unit) if isinstance(arg, u.Quantity)
            else (arg << u.dimensionless_unscaled).to_value(unit)
            for arg in args
        )
        kwargs = {
            k: kwargs[k].to_value(unit) if isinstance(kwargs[k], u.Quantity)
            else (kwargs[k] << u.dimensionless_unscaled).to_value(unit)
            for k in kwargs
        }

    if seed is None:
        func = getattr(np.random, func.__name__)
    else:
        func = getattr(np.random.default_rng(seed), func.__name__)

    value = func(
        *args,
        size=tuple(shape.values()),
        **kwargs,
    )

    if unit is not None:
        value = value << unit

    return na.ScalarArray(
        ndarray=value,
        axes=tuple(shape.keys()),
    )


@_implements(na.random.binomial)
def random_binomial(
    n: int | u.Quantity | na.AbstractScalarArray,
    p: float | na.AbstractScalarArray,
    shape_random: None | dict[str, int] = None,
    seed: None | int = None,
):
    try:
        n = scalars._normalize(n)
        p = scalars._normalize(p)
    except na.ScalarTypeError:
        return NotImplemented

    if shape_random is None:
        shape_random = dict()

    shape_base = na.shape_broadcasted(n, p)
    shape = na.broadcast_shapes(shape_base, shape_random)

    n = n.ndarray_aligned(shape)
    p = p.ndarray_aligned(shape)

    unit = na.unit(n)

    if unit is not None:
        n = n.value

    if seed is None:
        func = np.random.binomial
    else:
        func = np.random.default_rng(seed).binomial

    value = func(
        n=n,
        p=p,
        size=tuple(shape.values()),
    )

    if unit is not None:
        value = value << unit

    return na.ScalarArray(
        ndarray=value,
        axes=tuple(shape.keys()),
    )


@_implements(na.random.gamma)
def random_gamma(
    shape: float | na.AbstractScalarArray,
    scale: float | u.Quantity | na.AbstractScalarArray = 1,
    shape_random: None | dict[str, int] = None,
    seed: None | int = None,
) -> na.ScalarArray:
    alpha = shape
    theta = scale

    try:
        alpha = scalars._normalize(alpha)
        theta = scalars._normalize(theta)
    except na.ScalarTypeError:
        return NotImplemented

    if shape_random is None:
        shape_random = dict()

    shape_base = na.shape_broadcasted(alpha, theta)
    shape = na.broadcast_shapes(shape_base, shape_random)

    alpha = alpha.ndarray_aligned(shape)
    theta = theta.ndarray_aligned(shape)

    unit = na.unit(theta)

    if unit is not None:
        theta = theta.value

    if seed is None:
        func = np.random.gamma
    else:
        func = np.random.default_rng(seed).gamma

    value = func(
        shape=alpha,
        scale=theta,
        size=tuple(shape.values()),
    )

    if unit is not None:
        value = value << unit

    return na.ScalarArray(
        ndarray=value,
        axes=tuple(shape.keys()),
    )


def plt_plot_like(
        func: Callable,
        *args: na.AbstractScalarArray,
        ax: None | matplotlib.axes.Axes | na.ScalarArray[npt.NDArray[matplotlib.axes.Axes]] = None,
        axis: None | str = None,
        where: bool | na.AbstractScalarArray = True,
        components: None | tuple[str, ...] = None,
        **kwargs,
) -> na.ScalarArray[npt.NDArray[None | matplotlib.artist.Artist]]:

    if components is not None:
        raise ValueError(f"`components` should be `None` for scalars, got {components}")

    try:
        args = tuple(scalars._normalize(arg) for arg in args)
        where = scalars._normalize(where)
        kwargs = {k: scalars._normalize(kwargs[k]) for k in kwargs}
    except na.ScalarTypeError:
        return NotImplemented

    if ax is None:
        ax = plt.gca()
    ax = na.as_named_array(ax)

    shape_args = na.shape_broadcasted(*args)

    shape = na.broadcast_shapes(ax.shape, shape_args)

    if axis is None:
        if len(shape_args) != 1:
            raise ValueError(
                f"if `axis` is `None`, the broadcasted shape of `*args`, "
                f"{shape_args}, should have one element"
            )
        axis = next(iter(shape_args))

    shape_orthogonal = {a: shape[a] for a in shape if a != axis}

    args = tuple(arg.broadcast_to(shape) for arg in args)

    if not set(ax.shape).issubset(shape_orthogonal):
        raise ValueError(
            f"the shape of `ax`, {ax.shape}, "
            f"should be a subset of the broadcasted shape of `*args` excluding `axis`, {shape_orthogonal}",
        )
    ax = ax.broadcast_to(shape_orthogonal)

    if not set(where.shape).issubset(shape_orthogonal):
        raise ValueError(
            f"the shape of `where`, {where.shape}, "
            f"should be a subset of the broadcasted shape of `*args` excluding `axis`, {shape_orthogonal}"
        )
    where = where.broadcast_to(shape_orthogonal)

    kwargs_broadcasted = dict()
    for k in kwargs:
        kwarg = kwargs[k]
        if not set(na.shape(kwarg)).issubset(shape_orthogonal):
            raise ValueError(
                f"the shape of `{k}`, {na.shape(kwarg)}, "
                f"should be a subset of the broadcasted shape of `*args` excluding `axis`, {shape_orthogonal}"
            )
        kwargs_broadcasted[k] = na.broadcast_to(kwarg, shape_orthogonal)
    kwargs = kwargs_broadcasted

    result = na.ScalarArray.empty(shape=shape_orthogonal, dtype=object)

    for index in na.ndindex(shape_orthogonal):
        if where[index]:
            func_matplotlib = getattr(ax[index].ndarray, func.__name__)
            args_index = tuple(arg[index].ndarray for arg in args)
            kwargs_index = {k: kwargs[k][index].ndarray for k in kwargs}
            result[index] = func_matplotlib(
                *args_index,
                **kwargs_index,
            )[0]

    return result


@_implements(na.plt.scatter)
def plt_scatter(
        *args: na.AbstractScalarArray,
        s: None | na.AbstractScalarArray = None,
        c: None | na.AbstractScalarArray = None,
        ax: None | matplotlib.axes.Axes | na.ScalarArray = None,
        where: bool | na.AbstractScalarArray = True,
        components: None | tuple[str, ...] = None,
        **kwargs,
) -> na.ScalarArray:

    if components is not None:
        raise ValueError(f"`components` should be `None` for scalars, got {components}")

    try:
        args = tuple(scalars._normalize(arg) for arg in args)
        s = scalars._normalize(s)
        c = scalars._normalize(c)
        where = scalars._normalize(where)
        kwargs = {k: scalars._normalize(kwargs[k]) for k in kwargs}
    except na.ScalarTypeError:
        return NotImplemented

    if ax is None:
        ax = plt.gca()
    ax = na.as_named_array(ax)

    shape_c = c.shape
    if "rgba" in c.shape:
        shape_c.pop("rgba")

    shape = na.shape_broadcasted(*args, s, ax, where)
    shape = na.broadcast_shapes(shape, shape_c)

    shape_orthogonal = ax.shape

    args = tuple(arg.broadcast_to(shape) for arg in args)

    if np.issubdtype(na.get_dtype(s), np.number):
        s = na.broadcast_to(s, shape)
    else:
        s = na.broadcast_to(s, shape_orthogonal)

    if np.issubdtype(na.get_dtype(c), np.number):
        if "rgba" in c.shape:
            c = na.broadcast_to(c, shape | dict(rgba=c.shape["rgba"]))
        else:
            c = na.broadcast_to(c, shape)
    else:
        c = na.broadcast_to(c, shape_orthogonal)

    where = where.broadcast_to(shape)

    args = tuple(np.where(where, arg, np.nan) for arg in args)

    kwargs_broadcasted = dict()
    for k in kwargs:
        kwarg = kwargs[k]
        if not set(na.shape(kwarg)).issubset(shape_orthogonal):
            raise ValueError(
                f"the shape of `{k}`, {na.shape(kwarg)}, "
                f"should be a subset of the shape of `ax`, {shape_orthogonal}"
            )
        kwargs_broadcasted[k] = na.broadcast_to(kwarg, shape_orthogonal)
    kwargs = kwargs_broadcasted

    result = na.ScalarArray.empty(shape=shape_orthogonal, dtype=object)

    for index in na.ndindex(shape_orthogonal):
        func_matplotlib = getattr(ax[index].ndarray, "scatter")
        args_index = tuple(arg[index].ndarray.reshape(-1) for arg in args)

        s_index = s[index].ndarray
        if s_index is not None:
            s_index = s_index.reshape(-1)

        c_index = c[index].ndarray
        if c_index is not None:
            if "rgba" in c.shape:
                c_index = c[index].ndarray.reshape(-1, c.shape["rgba"])
            else:
                c_index = c[index].ndarray.reshape(-1)

        kwargs_index = {k: kwargs[k][index].ndarray for k in kwargs}
        result[index] = func_matplotlib(
            *args_index,
            s=s_index,
            c=c_index,
            **kwargs_index,
        )

    return result


@_implements(na.plt.stairs)
def plt_stairs(
        *args: na.AbstractScalarArray,
        ax: None | matplotlib.axes.Axes | na.ScalarArray[npt.NDArray[matplotlib.axes.Axes]] = None,
        axis: None | str = None,
        where: bool | na.AbstractScalarArray = True,
        **kwargs,
) -> na.ScalarArray[npt.NDArray[None | matplotlib.artist.Artist]]:

    if len(args) == 1:
        edges = None
        values, = args
    elif len(args) == 2:
        edges, values = args
    else:   # pragma: nocover
        raise ValueError(
            f"incorrect number of arguments, expected 1 or 2, got {len(args)}"
        )

    try:
        values = scalars._normalize(values)
        edges = scalars._normalize(edges) if edges is not None else edges
        where = scalars._normalize(where)
        kwargs = {k: scalars._normalize(kwargs[k]) for k in kwargs}
    except na.ScalarTypeError:  # pragma: nocover
        return NotImplemented

    if ax is None:
        ax = plt.gca()
    ax = na.as_named_array(ax)

    if axis is None:
        if len(values.shape) != 1:
            raise ValueError(
                f"if {axis=}, {values.shape=} should have only one element."
            )
        axis = next(iter(values.shape))
    else:
        if axis not in values.shape:
            raise ValueError(
                f"{axis=} must be an element of {values.shape}"
            )

    shape_values = na.shape(values)
    shape_edges = na.shape(edges)
    shape_args = na.broadcast_shapes(
        shape_values,
        {a: shape_edges[a] for a in shape_edges if a != axis},
    )

    shape = na.broadcast_shapes(ax.shape, shape_args)

    shape_orthogonal = {a: shape[a] for a in shape if a != axis}

    values = na.broadcast_to(values, shape)

    if edges is not None:
        edges = na.broadcast_to(
            array=edges,
            shape=shape_orthogonal | {axis: shape[axis] + 1},
        )

    if not set(ax.shape).issubset(shape_orthogonal):    # pragma: nocover
        raise ValueError(
            f"the shape of `ax`, {ax.shape}, "
            f"should be a subset of the broadcasted shape of `*args` excluding `axis`, {shape_orthogonal}",
        )
    ax = ax.broadcast_to(shape_orthogonal)

    if not set(where.shape).issubset(shape_orthogonal):     # pragma: nocover
        raise ValueError(
            f"the shape of `where`, {where.shape}, "
            f"should be a subset of the broadcasted shape of `*args` excluding `axis`, {shape_orthogonal}"
        )
    where = where.broadcast_to(shape_orthogonal)

    kwargs_broadcasted = dict()
    for k in kwargs:
        kwarg = kwargs[k]
        if not set(na.shape(kwarg)).issubset(shape_orthogonal):     # pragma: nocover
            raise ValueError(
                f"the shape of `{k}`, {na.shape(kwarg)}, "
                f"should be a subset of the broadcasted shape of `*args` excluding `axis`, {shape_orthogonal}"
            )
        kwargs_broadcasted[k] = na.broadcast_to(kwarg, shape_orthogonal)
    kwargs = kwargs_broadcasted

    result = na.ScalarArray.empty(shape=shape_orthogonal, dtype=object)

    for index in na.ndindex(shape_orthogonal):
        if where[index]:
            values_index = values[index].ndarray
            edges_index = edges[index].ndarray if edges is not None else edges
            kwargs_index = {k: kwargs[k][index].ndarray for k in kwargs}
            result[index] = ax[index].ndarray.stairs(
                values=values_index,
                edges=edges_index,
                **kwargs_index,
            )

    return result


@_implements(na.plt.imshow)
def plt_imshow(
    X: na.AbstractScalarArray,
    axis_x: str,
    axis_y: str,
    axis_rgb: None | str = None,
    ax: None | matplotlib.axes.Axes | na.AbstractArray = None,
    cmap: None | str | matplotlib.colors.Colormap = None,
    norm: None | str | matplotlib.colors.Normalize = None,
    aspect: None | na.ArrayLike = None,
    alpha: None | na.ArrayLike = None,
    vmin: None | na.ArrayLike = None,
    vmax: None | na.ArrayLike = None,
    extent: None | na.ArrayLike = None,
    **kwargs,
) -> na.ScalarArray:
    try:
        X = scalars._normalize(X)
        aspect = scalars._normalize(aspect) if aspect is not None else aspect
        alpha = scalars._normalize(alpha) if alpha is not None else alpha
        vmin = scalars._normalize(vmin) if vmin is not None else vmin
        vmax = scalars._normalize(vmax) if vmax is not None else vmax
        extent = scalars._normalize(extent) if extent is not None else extent
    except na.ScalarTypeError:  # pragma: nocover
        return NotImplemented

    if ax is None:
        ax = plt.gca()
    ax = na.as_named_array(ax)

    if axis_x not in X.shape:   # pragma: nocover
        raise ValueError(f"`{axis_x=}` must be a member of `{X.shape=}`")

    if axis_y not in X.shape:   # pragma: nocover
        raise ValueError(f"`{axis_y=}` must be a member of `{X.shape=}`")

    if axis_rgb is not None:    # pragma: nocover
        if axis_rgb not in X.shape:
            raise ValueError(f"`{axis_rgb=}` must be a member of `{X.shape=}`")

    shape_X = na.shape_broadcasted(X, ax)
    shape = {a: shape_X[a] for a in shape_X if a != axis_rgb}
    shape_orthogonal = ax.shape
    shape_extent = shape_orthogonal | {f"{axis_x},{axis_y}": 4}

    X = X.broadcast_to(shape_X)
    aspect = aspect.broadcast_to(shape_orthogonal) if aspect is not None else aspect
    alpha = alpha.broadcast_to(shape) if alpha is not None else alpha
    vmin = vmin.broadcast_to(shape_orthogonal) if vmin is not None else vmin
    vmax = vmax.broadcast_to(shape_orthogonal) if vmax is not None else vmax
    extent = extent.broadcast_to(shape_extent) if extent is not None else extent

    shape_index = {axis_y: shape[axis_y], axis_x: shape[axis_x]}

    if axis_rgb is not None:
        shape_X_index = shape_index | {axis_rgb: shape_X[axis_rgb]}
    else:
        shape_X_index = shape_index

    result = na.ScalarArray.empty(shape_orthogonal, dtype=object)

    for index in na.ndindex(shape_orthogonal):
        result[index] = ax[index].ndarray.imshow(
            X=X[index].ndarray_aligned(shape_X_index),
            cmap=cmap,
            norm=norm,
            aspect=aspect[index].ndarray if aspect is not None else aspect,
            alpha=alpha[index].ndarray_aligned(shape_index) if alpha is not None else alpha,
            vmin=vmin[index].ndarray if vmin is not None else vmin,
            vmax=vmax[index].ndarray if vmax is not None else vmax,
            extent=extent[index].ndarray if extent is not None else extent,
            **kwargs,
        )

    return result


@_implements(na.plt.pcolormesh)
def pcolormesh(
    *XY: na.AbstractScalarArray,
    C: na.AbstractScalarArray,
    axis_rgb: None | str = None,
    ax: None | matplotlib.axes.Axes | na.AbstractScalarArray = None,
    components: None | tuple[str, str] = None,
    cmap: None | str | matplotlib.colors.Colormap = None,
    norm: None | str | matplotlib.colors.Normalize = None,
    vmin: None | float | u.Quantity | na.AbstractScalarArray = None,
    vmax: None | float | u.Quantity | na.AbstractScalarArray = None,
    **kwargs,
) -> na.ScalarArray:

    if components is not None:  # pragma: nocover
        raise ValueError(f"`components` should be `None` for scalars, got {components}")

    try:
        XY = tuple(scalars._normalize(arg) for arg in XY)
        C = scalars._normalize(C)
        vmin = scalars._normalize(vmin) if vmin is not None else vmin
        vmax = scalars._normalize(vmax) if vmax is not None else vmax
    except na.ScalarTypeError:  # pragma: nocover
        pass

    if ax is None:
        ax = plt.gca()
    ax = na.as_named_array(ax)

    if axis_rgb is not None:    # pragma: nocover
        if axis_rgb not in C.shape:
            raise ValueError(f"`{axis_rgb=}` must be a member of `{C.shape=}`")

    shape_orthogonal = ax.shape

    shape_XY = na.shape_broadcasted(ax, *XY)
    shape_C = na.shape_broadcasted(ax, C)

    axes_XY = tuple(ax for ax in shape_C if ax not in shape_orthogonal)
    axes_XY = tuple(ax for ax in axes_XY if ax != axis_rgb)

    axes_C = axes_XY
    if axis_rgb is not None:
        axes_C = axes_C + (axis_rgb,)

    XY = tuple(arg.broadcast_to(shape_XY) for arg in XY)
    C = C.broadcast_to(shape_C)
    vmin = vmin.broadcast_to(shape_orthogonal) if vmin is not None else vmin
    vmax = vmax.broadcast_to(shape_orthogonal) if vmax is not None else vmax

    result = na.ScalarArray.empty(shape_orthogonal, dtype=object)

    for index in na.ndindex(shape_orthogonal):
        result[index] = ax[index].ndarray.pcolormesh(
            *[arg[index].ndarray_aligned(axes_XY) for arg in XY],
            C[index].ndarray_aligned(axes_C),
            cmap=cmap,
            norm=norm,
            vmin=vmin[index].ndarray if vmin is not None else vmin,
            vmax=vmax[index].ndarray if vmax is not None else vmax,
            **kwargs,
        )

    return result


@_implements(na.plt.pcolormovie)
def pcolormovie(
    *TXY: na.AbstractScalarArray,
    C: na.AbstractScalarArray,
    axis_time: str,
    axis_rgb: None | str = None,
    ax: None | matplotlib.axes.Axes | na.AbstractScalarArray = None,
    components: None | tuple[str, str] = None,
    cmap: None | str | matplotlib.colors.Colormap = None,
    norm: None | str | matplotlib.colors.Normalize = None,
    vmin: None | float | u.Quantity | na.AbstractScalarArray = None,
    vmax: None | float | u.Quantity | na.AbstractScalarArray = None,
    kwargs_pcolormesh: None | dict[str, Any] = None,
    kwargs_animation: None | dict[str, Any] = None,
) -> matplotlib.animation.FuncAnimation:

    t, x, y = TXY

    if ax is None:
        ax = plt.gca()
    ax = na.asanyarray(ax)

    ax0 = ax.ndarray.flat[0]
    fig = ax0.figure

    try:
        t = scalars._normalize(t)
        x = scalars._normalize(x)
        y = scalars._normalize(y)
        C = scalars._normalize(C)
    except na.ScalarTypeError:  # pragma: nocover
        return NotImplemented

    shape = na.shape_broadcasted(t, x, y)
    t = t.broadcast_to(na.shape_broadcasted(t, ax))
    x = x.broadcast_to(shape)
    y = y.broadcast_to(shape)

    if kwargs_pcolormesh is None:
        kwargs_pcolormesh = dict()
    if kwargs_animation is None:
        kwargs_animation = dict()

    def func(frame: int):
        index_frame = {axis_time: frame}
        for i in ax.ndindex():
            ax_i = ax[i].ndarray
            for artist in ax_i.collections:
                artist.remove()
            ax_i.relim()
            ax_i.set_title(t[index_frame | i].ndarray)

        na.plt.pcolormesh(
            x[index_frame],
            y[index_frame],
            C=C[index_frame],
            axis_rgb=axis_rgb,
            ax=ax,
            components=components,
            cmap=cmap,
            norm=norm,
            vmin=vmin,
            vmax=vmax,
            **kwargs_pcolormesh,
        )

    func(0)

    result = matplotlib.animation.FuncAnimation(
        fig=fig,
        func=func,
        frames=shape[axis_time],
        **kwargs_animation,
    )

    return result


@_implements(na.plt.text)
def plt_text(
    x: float | u.Quantity | na.AbstractScalarArray,
    y: float | u.Quantity | na.AbstractScalarArray,
    s: str | na.AbstractScalarArray,
    ax: None | matplotlib.axes.Axes | na.AbstractScalarArray = None,
    **kwargs,
) -> na.AbstractScalarArray:

    if ax is None:
        ax = plt.gca()

    try:
        x = scalars._normalize(x)
        y = scalars._normalize(y)
        s = scalars._normalize(s)
        ax = scalars._normalize(ax)
        kwargs = {k: scalars._normalize(kwargs[k]) for k in kwargs}
    except na.ScalarTypeError:  # pragma: nocover
        return NotImplemented

    shape = na.shape_broadcasted(x, y, s, ax)

    x = x.broadcast_to(shape)
    y = y.broadcast_to(shape)
    s = s.broadcast_to(shape)
    ax = ax.broadcast_to(shape)
    kwargs = {k: kwargs[k].broadcast_to(shape) for k in kwargs}

    result = na.ScalarArray.empty(shape, dtype=matplotlib.axes.Axes)

    for index in na.ndindex(shape):
        kwargs_index = {k: kwargs[k][index].ndarray for k in kwargs}
        result[index] = ax[index].ndarray.text(
            x=x[index].ndarray,
            y=y[index].ndarray,
            s=s[index].ndarray,
            **kwargs_index,
        )

    return result


def plt_axes_setter(
    method: str,
    *args,
    ax: None | matplotlib.axes.Axes | na.AbstractScalarArray = None,
    **kwargs,
) -> None:

    if ax is None:
        ax = plt.gca()

    try:
        args = [scalars._normalize(arg) for arg in args]
        ax = scalars._normalize(ax)
        kwargs = {k: scalars._normalize(kwargs[k]) for k in kwargs}
    except na.ScalarTypeError:  # pragma: nocover
        return NotImplemented

    shape = ax.shape

    args = [arg.broadcast_to(shape) for arg in args]
    kwargs = {k: kwargs[k].broadcast_to(shape) for k in kwargs}

    for index in na.ndindex(shape):
        args_index = [arg[index].ndarray for arg in args]
        kwargs_index = {k: kwargs[k][index].ndarray for k in kwargs}
        getattr(ax[index].ndarray, method.__name__)(*args_index, **kwargs_index)


def plt_axes_getter(
    method: str,
    ax: na.AbstractScalarArray,
) -> na.ScalarArray:

    try:
        ax = scalars._normalize(ax)
    except na.ScalarTypeError:  # pragma: nocover
        return NotImplemented

    result = na.ScalarArray.empty(shape=ax.shape, dtype=object)

    for index in na.ndindex(ax.shape):
        ax_index = ax[index].ndarray
        if ax_index is None:
            ax_index = plt.gca()
        result[index] = getattr(ax_index, method.__name__)()

    return result


def plt_axes_attribute(
    method: str,
    ax: na.AbstractScalarArray,
) -> na.ScalarArray:

    try:
        ax = scalars._normalize(ax)
    except na.ScalarTypeError:  # pragma: nocover
        return NotImplemented

    result = na.ScalarArray.empty(shape=ax.shape, dtype=object)

    for index in na.ndindex(ax.shape):
        ax_index = ax[index].ndarray
        if ax_index is None:
            ax_index = plt.gca()
        result[index] = getattr(ax_index, method.__name__)

    return result


@_implements(na.jacobian)
def jacobian(
        function: Callable[[na.AbstractScalar], na.AbstractScalar],
        x: na.AbstractScalar,
        dx: None | na.AbstractScalar = None,
        like: None | na.AbstractScalar = None,
) -> na.AbstractScalar:

    f = function(x)

    if isinstance(f, na.AbstractScalar):
        if isinstance(x, na.AbstractScalar):
            x0 = x + dx
            f0 = function(x0)
            df = f0 - f
            return df / dx

        else:
            return NotImplemented
    else:
        return NotImplemented


@_implements(na.optimize.root_newton)
def optimize_root_newton(
        function: Callable[[na.ScalarLike], na.ScalarLike],
        guess: na.ScalarLike,
        jacobian: Callable[[na.ScalarLike], na.ScalarLike],
        max_abs_error: na.ScalarLike,
        max_iterations: int = 100,
        callback: None | Callable[[int, na.ScalarLike, na.ScalarLike, na.ScalarLike], None] = None,
) -> na.ScalarArray:

    if isinstance(guess, na.AbstractArray):
        if isinstance(guess, na.AbstractScalar):
            pass
        else:
            return NotImplemented
    else:
        guess = na.ScalarArray(guess)

    x = guess
    f = function(x)

    if isinstance(f, na.AbstractArray):
        if isinstance(f, na.AbstractScalar):
            pass
        else:
            return NotImplemented
    else:
        f = na.ScalarArray(f)

    if isinstance(max_abs_error, na.AbstractArray):
        if isinstance(max_abs_error, na.AbstractScalar):
            pass
        else:
            return NotImplemented
    else:
        max_abs_error = na.ScalarArray(max_abs_error)

    if na.shape(max_abs_error):
        raise ValueError(f"argument `max_abs_error` should have an empty shape, got {na.shape(max_abs_error)}")

    shape = na.shape_broadcasted(f, guess)

    converged = na.broadcast_to(0 * na.value(f), shape=shape).astype(bool)

    x = na.broadcast_to(x, shape).astype(float)

    for i in range(max_iterations):

        if callback is not None:
            callback(i, x, f, converged)

        converged = np.abs(f) < max_abs_error

        if np.all(converged):
            return x

        jac = jacobian(x)

        correction = f / jac

        x = x - correction

        f = function(x)

    raise ValueError("Max iterations exceeded")


@_implements(na.optimize.root_secant)
def optimize_root_secant(
        function: Callable[[na.ScalarLike], na.ScalarLike],
        guess: na.ScalarLike,
        min_step_size: na.ScalarLike,
        max_abs_error: na.ScalarLike,
        max_iterations: int = 100,
        damping: None | float = None,
        callback: None | Callable[[int, na.ScalarLike, na.ScalarLike, na.ScalarLike], None] = None,
) -> na.ScalarArray:

    try:
        guess = scalars._normalize(guess)

        min_step_size = scalars._normalize(min_step_size)

        x0 = guess - 10 * min_step_size
        x1 = guess

        f0 = function(x0)
        f0 = scalars._normalize(f0)

        max_abs_error = scalars._normalize(max_abs_error)

    except scalars.ScalarTypeError:
        return NotImplemented

    if na.shape(max_abs_error):
        raise ValueError(f"argument `max_abs_error` should have an empty shape, got {na.shape(max_abs_error)}")

    shape = na.shape_broadcasted(f0, guess, min_step_size)

    converged = na.broadcast_to(0 * na.value(f0), shape=shape).astype(bool)

    x1 = na.broadcast_to(x1, shape).astype(float)

    for i in range(max_iterations):


        f1 = function(x1)

        if callback is not None:
            callback(i, x1, f1, converged)

        if max_abs_error is not None:
            converged |= np.abs(f1) < max_abs_error

        dx = x1 - x0

        converged |= np.abs(dx) < np.abs(min_step_size)

        if np.all(converged):
            return x1

        active = ~converged

        df = f1 - f0

        if np.any(df == 0, where=active):
            raise ValueError("stationary point detected")

        dx_active = dx[active]
        f0_active = f0[active]
        f1_active = f1[active]

        df_active = f1_active - f0_active
        if np.any(df_active == 0):
            raise ValueError("stationary point detected")

        jacobian = df_active / dx_active

        correction = f1_active / jacobian
        if damping is not None:
            correction = damping * correction

        x2 = x1.copy()
        x2[active] -= correction

        x0 = x1
        x1 = x2
        f0 = f1

    raise ValueError("Max iterations exceeded")


@_implements(na.colorsynth.rgb)
def colorsynth_rgb(
    spd: na.AbstractScalarArray,
    wavelength: None | na.AbstractScalarArray = None,
    axis: None | str = None,
    spd_min: None | float | u.Quantity | na.AbstractScalarArray = None,
    spd_max: None | float | u.Quantity | na.AbstractScalarArray = None,
    spd_norm: None | Callable = None,
    wavelength_min: None | float | u.Quantity | na.AbstractScalarArray = None,
    wavelength_max: None | float | u.Quantity | na.AbstractScalarArray = None,
    wavelength_norm: None | Callable = None,
) -> na.ScalarArray:
    try:
        spd = scalars._normalize(spd).astype(float)
        wavelength = scalars._normalize(wavelength) if wavelength is not None else wavelength
        spd_min = scalars._normalize(spd_min) if spd_min is not None else spd_min
        spd_max = scalars._normalize(spd_max) if spd_max is not None else spd_max
        wavelength_min = scalars._normalize(wavelength_min) if wavelength_min is not None else wavelength_min
        wavelength_max = scalars._normalize(wavelength_max) if wavelength_max is not None else wavelength_max
    except na.ScalarTypeError:  # pragma: nocover
        return NotImplemented

    if axis is None:
        axes = tuple(set(na.shape(spd)) | set(na.shape(wavelength)))
        if len(axes) != 1:
            raise ValueError(
                f"If `axis` is `None`, the other arguments must have zero"
                f"or one axis, got {axes}."
            )
        axis = axes[0]

    if wavelength is not None:
        if axis in spd.shape:
            if wavelength.shape[axis] == spd.shape[axis] + 1:
                below = {axis: slice(None, ~0)}
                above = {axis: slice(+1, None)}
                wavelength = (wavelength[below] + wavelength[above]) / 2

    shape = na.shape_broadcasted(
        spd,
        wavelength,
        spd_min,
        spd_max,
        wavelength_min,
        wavelength_max,
    )

    axes = tuple(shape)
    axis_ndarray = axes.index(axis)

    result_ndarray = colorsynth.rgb(
        spd=spd.ndarray_aligned(shape),
        wavelength=wavelength.ndarray_aligned(shape) if wavelength is not None else wavelength,
        axis=axis_ndarray,
        spd_min=spd_min.ndarray_aligned(shape) if spd_min is not None else spd_min,
        spd_max=spd_max.ndarray_aligned(shape) if spd_max is not None else spd_max,
        spd_norm=spd_norm,
        wavelength_min=wavelength_min.ndarray_aligned(shape) if wavelength_min is not None else wavelength_min,
        wavelength_max=wavelength_max.ndarray_aligned(shape) if wavelength_max is not None else wavelength_max,
        wavelength_norm=wavelength_norm,
    )

    result = na.ScalarArray(
        ndarray=result_ndarray,
        axes=axes,
    )

    return result


@_implements(na.colorsynth.colorbar)
def colorsynth_colorbar(
    spd: na.AbstractScalarArray,
    wavelength: None | na.AbstractScalarArray = None,
    axis: None | str = None,
    axis_wavelength: str = "_wavelength",
    axis_intensity: str = "_intensity",
    spd_min: None | float | u.Quantity | na.AbstractScalarArray = None,
    spd_max: None | float | u.Quantity | na.AbstractScalarArray = None,
    spd_norm: None | Callable = None,
    wavelength_min: None | float | u.Quantity | na.AbstractScalarArray = None,
    wavelength_max: None | float | u.Quantity | na.AbstractScalarArray = None,
    wavelength_norm: None | Callable = None,
) -> na.FunctionArray[na.Cartesian2dVectorArray, na.ScalarArray]:
    try:
        spd = scalars._normalize(spd).astype(float)
        wavelength = scalars._normalize(wavelength) if wavelength is not None else wavelength
        spd_min = scalars._normalize(spd_min) if spd_min is not None else spd_min
        spd_max = scalars._normalize(spd_max) if spd_max is not None else spd_max
        wavelength_min = scalars._normalize(wavelength_min) if wavelength_min is not None else wavelength_min
        wavelength_max = scalars._normalize(wavelength_max) if wavelength_max is not None else wavelength_max
    except na.ScalarTypeError:  # pragma: nocover
        return NotImplemented

    if axis is None:
        axes = tuple(set(na.shape(spd)) | set(na.shape(wavelength)))
        if len(axes) != 1:
            raise ValueError(
                f"If `axis` is `None`, the other arguments must have zero "
                f"or one axis, got {axes}."
            )
        axis = axes[0]

    if wavelength is not None:
        if axis in spd.shape:
            if wavelength.shape[axis] == spd.shape[axis] + 1:
                below = {axis: slice(None, ~0)}
                above = {axis: slice(+1, None)}
                wavelength = (wavelength[below] + wavelength[above]) / 2

    shape = na.shape_broadcasted(
        spd,
        wavelength,
        spd_min,
        spd_max,
        wavelength_min,
        wavelength_max,
    )

    axes = tuple(shape)
    axis_ndarray = axes.index(axis)

    intensity, wavelength, rgb = colorsynth.colorbar(
        spd=spd.ndarray_aligned(shape),
        wavelength=wavelength.ndarray_aligned(shape) if wavelength is not None else wavelength,
        axis=axis_ndarray,
        spd_min=spd_min.ndarray_aligned(shape) if spd_min is not None else spd_min,
        spd_max=spd_max.ndarray_aligned(shape) if spd_max is not None else spd_max,
        spd_norm=spd_norm,
        wavelength_min=wavelength_min.ndarray_aligned(shape) if wavelength_min is not None else wavelength_min,
        wavelength_max=wavelength_max.ndarray_aligned(shape) if wavelength_max is not None else wavelength_max,
        wavelength_norm=wavelength_norm,
        squeeze=False,
    )

    axes_new = (axis_wavelength, axis_intensity) + axes

    intensity = na.ScalarArray(intensity, axes_new)
    wavelength = na.ScalarArray(wavelength, axes_new)
    rgb = na.ScalarArray(rgb, axes_new)

    shape_intensity = intensity.shape
    shape_wavelength = wavelength.shape
    shape_rgb = rgb.shape

    intensity = intensity[{ax: 0 for ax in shape_intensity if shape_intensity[ax] == 1}]
    wavelength = wavelength[{ax: 0 for ax in shape_wavelength if shape_wavelength[ax] == 1}]
    rgb = rgb[{ax: 0 for ax in shape_rgb if shape_rgb[ax] == 1}]

    return na.FunctionArray(
        inputs=na.Cartesian2dVectorArray(
            x=intensity,
            y=wavelength,
        ),
        outputs=rgb,
    )


def ndfilter(
    func: Callable,
    array: na.AbstractScalarArray,
    size: dict[str, int],
    where: bool | na.AbstractScalarArray,
    **kwargs,
) -> na.ScalarArray:

    func = getattr(ndfilters, func.__name__)

    try:
        array = scalars._normalize(array)
        where = scalars._normalize(where)
    except scalars.ScalarTypeError:   # pragma: nocover
        return NotImplemented

    shape = na.shape_broadcasted(array, where)
    axes = tuple(shape)
    if not set(size).issubset(axes):
        raise ValueError(
            f"the keys in {size=} must be a subset of the keys in {shape=}."
        )

    array = array.broadcast_to(shape)
    where = where.broadcast_to(shape)

    return array.type_explicit(
        ndarray=func(
            array=array.ndarray,
            size=tuple(size.values()),
            axis=tuple(axes.index(ax) for ax in size),
            where=where.ndarray,
            **kwargs,
        ),
        axes=axes,
    )


@_implements(na.regridding.weights)
def regridding_weights(
    coordinates_input: na.AbstractScalarArray | na.AbstractVectorArray,
    coordinates_output: na.AbstractScalarArray | na.AbstractVectorArray,
    axis_input: None | str | Sequence[str] = None,
    axis_output: None | str | Sequence[str] = None,
    method: Literal['multilinear', 'conservative'] = 'multilinear',
) -> tuple[na.AbstractScalar, dict[str, int], dict[str, int]]:

    if not isinstance(coordinates_output, na.AbstractVectorArray):
        coordinates_output = na.CartesianNdVectorArray(dict(x=coordinates_output))

    if not isinstance(coordinates_input, na.AbstractVectorArray):
        coordinates_input = na.CartesianNdVectorArray(dict(x=coordinates_input))

    return na.regridding.weights(
        coordinates_input=coordinates_input,
        coordinates_output=coordinates_output,
        axis_input=axis_input,
        axis_output=axis_output,
        method=method,
    )


@_implements(na.regridding.regrid_from_weights)
def regridding_regrid_from_weights(
    weights: na.AbstractScalarArray,
    shape_input: dict[str, int],
    shape_output: dict[str, int],
    values_input: na.AbstractScalarArray,
) -> na.ScalarArray:

    try:
        weights = scalars._normalize(weights)
        values_input = scalars._normalize(values_input)
    except scalars.ScalarTypeError:  # pragma: nocover
        return NotImplemented

    shape_weights = weights.shape

    axis_input = tuple(a for a in shape_input if a not in shape_weights)
    axis_output = tuple(a for a in shape_output if a not in shape_weights)

    shape_values_input = values_input.shape
    shape_orthogonal = {
        a: shape_values_input[a]
        for a in shape_values_input
        if a not in axis_input
    }
    shape_orthogonal = na.broadcast_shapes(shape_orthogonal, shape_weights)

    shape_input = na.broadcast_shapes(shape_orthogonal, shape_input)
    shape_output = na.broadcast_shapes(shape_orthogonal, shape_output)

    weights = weights.broadcast_to({
        a: shape_input[a] if a not in axis_input else 1
        for a in shape_input
    })
    values_input = values_input.broadcast_to(shape_input)

    result = regridding.regrid_from_weights(
        weights=weights.ndarray,
        shape_input=tuple(shape_input.values()),
        shape_output=tuple(shape_output.values()),
        values_input=values_input.ndarray,
        axis_input=tuple(tuple(shape_input).index(a) for a in axis_input),
        axis_output=tuple(tuple(shape_output).index(a) for a in axis_output),
    )

    result = na.ScalarArray(
        ndarray=result,
        axes=tuple(shape_output),
    )

    return result


@_implements(na.despike)
def despike(
    array: na.AbstractScalarArray,
    axis: tuple[str, str],
    where: None | bool | na.AbstractScalarArray,
    inbkg: None | na.AbstractScalarArray,
    invar: None | float | na.AbstractScalarArray,
    sigclip: float,
    sigfrac: float,
    objlim: float,
    gain: float,
    readnoise: float,
    satlevel: float,
    niter: int,
    sepmed: bool,
    cleantype: Literal["median", "medmask", "meanmask", "idw"],
    fsmode: Literal["median", "convolve"],
    psfmodel: Literal["gauss", "gaussx", "gaussy", "moffat"],
    psffwhm: float,
    psfsize: int,
    psfk: None | na.AbstractScalarArray,
    psfbeta: float,
    verbose: bool,
) -> na.ScalarArray:

    try:
        array = scalars._normalize(array)
        where = scalars._normalize(where) if where is not None else where
        inbkg = scalars._normalize(inbkg) if inbkg is not None else inbkg
        invar = scalars._normalize(invar) if invar is not None else invar
        psfk = scalars._normalize(psfk) if psfk is not None else psfk
    except scalars.ScalarTypeError:  # pragma: nocover
        return NotImplemented

    shape = na.shape_broadcasted(
        array,
        where,
        inbkg,
        invar,
    )

    array = array.broadcast_to(shape)
    where = where.broadcast_to(shape) if where is not None else where
    inbkg = inbkg.broadcast_to(shape) if inbkg is not None else inbkg
    invar = invar.broadcast_to(shape) if invar is not None else invar

    if psfk is not None:
        shape_orthogonal = {ax: shape[ax] for ax in shape if ax not in axis}
        shape_psfk = na.broadcast_shapes(shape_orthogonal, psfk.shape)
        psfk = na.broadcast_to(psfk, shape_psfk)

    result = array.copy()
    inmask = ~where if where is not None else where

    for index in na.ndindex(shape, axis_ignored=axis):
        result_ndarray = astroscrappy.detect_cosmics(
            indat=array[index].ndarray,
            inmask=inmask[index].ndarray if inmask is not None else inmask,
            inbkg=inbkg[index].ndarray if inbkg is not None else inbkg,
            invar=invar[index].ndarray if invar is not None else invar,
            sigclip=sigclip,
            sigfrac=sigfrac,
            objlim=objlim,
            gain=gain,
            readnoise=readnoise,
            satlevel=satlevel,
            niter=niter,
            sepmed=sepmed,
            cleantype=cleantype,
            fsmode=fsmode,
            psfmodel=psfmodel,
            psffwhm=psffwhm,
            psfsize=psfsize,
            psfk=psfk[index].ndarray if psfk is not None else psfk,
            psfbeta=psfbeta,
            verbose=verbose,
        )

        result[index].value.ndarray[:] = result_ndarray[1]

    return result
